#!/usr/bin/python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""DielectricFunction.

Provides an interface to different mechanisms for providing dielectric information

- Base class :class:`DielectricFunction`

    - Subclass :class:`ConstantTensor`
    - Subclass :class:`ConstantScalar`
    - Subclass :class:`TabulateScalar`
    - Subclass :class:`Tabulate1`
    - Subclass :class:`Tabulate3`
    - Subclass :class:`Tabulate6`
    - Subclass :class:`DFT`
    - Subclass :class:`DrudeLorentz`
    - Subclass :class:`FPSQ`
    - Subclass :class:`Sellmeier`
"""

import sys
from contextlib import nullcontext

import numpy as np
from scipy import interpolate

from PDielec.Constants import angstrom, wavenumber


class DielectricFunction:
    """A base class to different dielectric functions.

    There is a close relationship between the dielectric function and a material.
    The Material module allows data to be read in for different dielectric models.
    Once the data has been read in a DielectricFunction is created which is appropriate for the data.
    The tensor and scalar versions of the functions behave differently with respect to 
    epsilon infinity: tensor functions add epsilon infinity to the calculated permittivity, scalar functions do not.

    Attributes
    ----------
    epsilon_infinity : ndarray
        A 3x3 tensor (numpy array) representing the optical permittivity.
    volume : float
        Volume of crystal unit cells.
    volume_au : float
        Volume of crystal unit cells in atomic units.
    vs_cm1 : ndarray
        An array of floats representing the frequencies of any tabulation.
    isScalarFunction : bool
        True if this function is a scalar function, False otherwise.

    Methods
    -------
    setVolume(volume)
        Change the unit cell volume.
    setEpsilonInfinity(eps)
        Change epsilon infinity.
    function()
        Return the calculate function for this object.
    calculate(v)
        Return the value of the permittivity at frequency v.
    setFrequencies(frequencies)
        Set frequencies 
    getFrequencies()
        Get frequencies 
    setSigmas(sigmas)
        Set sigmas 
    getSigmas()
        Get sigmas 
    setOscillatorStrengths(strengths)
        Set strengths 
    getOscillatorStrengths()
        Get strengths 
    dielectricContributionsFromDrude(f, frequency, sigma, volume)
        Calculates permittivity from a Drude Model.
    dielectricContributionsFromModes(f, modes, frequencies, sigmas, strengths, volume)
        Calculates permittivity from a Drude-Lorentz model.

    """

    possible_epsTypes = ["dft","fpsq","drude-lorentz","sellmeier"]

    def __init__(self ):
        """Initialise and instance of the function."""
        self.isScalarFunction             = None
        self.vs_cm1               = 0
        self.epsilon_infinity     = np.zeros( (3,3) )

    def setFrequencies(self, frequencies):
        """Set the frequencies of a Lorentzian permittivity.

        The default is to print an error message unless the DielectricFunction is a Lorentzian

        Parameters
        ----------
        frequencies : 1d array of floats
            The frequencies in cm-1

        Returns
        -------
        None

        """
        print("Error is DielectricFunction.setFrequencies: unable to set frequencies of a non-Lorentzian function")
        return 


    def getFrequencies(self):
        """Return the frequencies of a Lorentzian permittivity.

        The default is to return None, only return an array if the permittivity is Lorentzian

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        return


    def setOscillatorStrengths(self, strengths):
        """Set the oscillator strengths of a Lorentzian permittivity.

        The default is to print an error message unless the DielectricFunction is a Lorentzian

        Parameters
        ----------
        strengths : 1d array of floats
            The oscillator strengths in cm-1

        Returns
        -------
        None

        """
        print("Error is DielectricFunction.setOscillatorStrengths: unable to set oscillator strengths of a non-Lorentzian function")
        return 


    def getOscillatorStrengths(self):
        """Return the oscillator strengths of a Lorentzian permittivity.

        The default is to return None, only return an array if the permittivity is Lorentzian

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        return

    def setSigmas(self, sigmas):
        """Set the sigma parameters of a Lorentzian permittivity.

        The default is to print an error message unless the DielectricFunction is a Lorentzian

        Parameters
        ----------
        sigmas : 1d array of floats
            The sigma parameters in cm-1

        Returns
        -------
        None

        """
        print("Error is DielectricFunction.setSigmas: unable to set sigmas of a non-Lorentzian function")
        return 


    def getSigmas(self):
        """Return the sigma parameters of a Lorentzian permittivity.

        The default is to return None, only return an array if the permittivity is Lorentzian

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        return


    def isTensor(self):
        """Return true if the dielectric function is a tensor.

        Parameters
        ----------
        None

        Returns
        -------
        bool
            True if the dielectric function returns a tensor, otherwise False.

        """
        return not self.isScalarFunction

    def isScalar(self):
        """Return true if the dielectric function is a scalar.

        Parameters
        ----------
        None

        Returns
        -------
        bool
            True if the dielectric function returns a scalar value, False otherwise.

        """
        return self.isScalarFunction

    def setVolume(self,volume):
        """Set the volume for dielectric calculations.

        Parameters
        ----------
        volume : float
            The volume in angs^3.

        Returns
        -------
        None

        """
        self.volume_angs     = volume
        self.volume_au       = volume*angstrom*angstrom*angstrom
        return

    def setEpsilonInfinity(self,eps):
        """Set epsilon infinity for dielectric calculations.

        Parameters
        ----------
        eps : tensor (3x3)
            The value of eps_infinity.

        Returns
        -------
        None

        """
        self.epsilon_infinity     = np.array(eps)
        return

    def function(self):
        """Return the function used to calculate the permittivity.

        Parameters
        ----------
        None

        Returns
        -------
        a function

        """
        return  self.calculate

    def getLowestFrequency(self):
        """Return the lowest tabulated frequency in cm-1.

        Parameters
        ----------
        None

        Returns
        -------
        The lowest tabulated frequency in cm-1.

        """
        return np.min(self.vs_cm1)

    def getHighestFrequency(self):
        """Return the highest tabulated frequency in cm-1.

        Parameters
        ----------
        None

        Returns
        -------
        float
            The highest tabulated frequency in cm-1.

        """
        return np.max(self.vs_cm1)

    def dielectriContributionsFromDrude(self, f, frequency, sigma, volume):
        """Calculate the dielectric function for a set of a Drude oscillator.

        The units of the parameters must be self-consistent.  Atomic units is a good choice.

        Parameters
        ----------
        f : float
            The frequency of the dielectric response.
        frequency : float
            The frequency of the drude oscillator.
        sigma : float
            The width of the oscillator.
        volume : float
            The volume of the cell

        Returns
        -------
        tensor (3x3)
            A complex scalar dielectric.

        """
        dielectric = np.zeros((3, 3), dtype=complex)
        unit = np.eye(3,dtype=complex)
        # Avoid a divide by zero if f is small
        f = max(1.0e-8, f)
        # Assume that the drude contribution is isotropic
        dielectric = dielectric - unit * frequency*frequency / complex(-f*f, -sigma*f)
        return dielectric * (4.0*np.pi/volume)


    def dielectriContributionsFromModes(self, f, modes, frequencies, sigmas, strengths, volume):
        """Calculate the dielectric function for a set of modes.

        Parameters
        ----------
        f : float
            The frequency considered
        modes : list of int
            A list of the modes to be considered.
        frequencies : list of float
            The frequency of each mode.
        sigmas : list of float
            The width of each mode.
        strengths : list of float
            The strengths of each mode.
        volume : float
            The volume of the cell

        Returns
        -------
        tensor (3x3)
            A complex scalar representing the dielectric function.

        """
        dielectric = np.zeros((3, 3), dtype=complex)
        for mode in modes:
            v = frequencies[mode]
            sigma = sigmas[mode]
            strength = strengths[mode].astype(complex)
            dielectric = dielectric + strength / complex((v*v - f*f), -sigma*f)
        return dielectric * (4.0*np.pi/volume)

    def print(self, v1_cm1, v2_cm1, v_inc=1,diagonal_only=True,file=None):
        """Print the permittivity over a range of frequencies to a file.

        Parameters
        ----------
        v1_cm1 : float
            Starting frequency
        v2_cm1 : float
            Finishing frequency
        v_inc : float, optional
            Incrementing frequency, default is 1
        diagonal_only: bool,optional
            If true only the diagonal values are printed
        file : str, optional
            Filename for printing.  If 'None' then stdout is used

        Returns
        -------
        None

        """
        c = ","
        with open(file, "w") if file else nullcontext(sys.stdout) as fd:
            if diagonal_only:
                print("f,eps00.real,epss11.real,eps22.real,eps00.imag,eps11.imag,eps22,imag",file=fd)
            else:
                print("f,eps00,epss11,eps22,eps01,eps02,,eps12",file=fd)
            for v in np.arange(v1_cm1, v2_cm1, v_inc):
                epsilon = self.calculate(v)
                if diagonal_only:
                    print(v,c,epsilon[0,0].real,c,epsilon[1,1].real,c,epsilon[2,2].real,c,epsilon[0,0].imag,c,epsilon[1,1].imag,c,epsilon[2,2].imag,file=fd)
                else:
                    print(v,c,epsilon[0,0],c,epsilon[1,1],c,epsilon[2,2],c,epsilon[0,1],c,epsilon[0,2],c,epsilon[1,2],file=fd)
        return

class ConstantTensor(DielectricFunction):
    """A complex constant tensor permittivity.

    Inherits from DielectricFunction.
    Provides a calculate() function to return the permittivity.
    Epsilon infinity is used to calculate the total permittivity

    Methods
    -------
    calculate()
        Calculates and returns the permittivity.

    Examples
    --------
    .. code-block:: python

        permittivity = np.array([ [ 2.0, 0.0, 0.0 ] , [ 0.0, 3.0, 0.0 ], [ 0.0, 0.0, 4.0 ] ])
        df = ConstantTensor(permittivity)
        newMaterial = Material('New material name',permittivityObject=df,cell=UnitCell(2, 2, 2))

    """

    def __init__(self, value):
        """Initialise ConstantTensor.

        Parameters
        ----------
        value : 3x3 numpy array
            Initialise the permittivity with this tensor

        """
        DielectricFunction.__init__(self)
        self.value = value
        self.isScalarFunction = False

    def calculate(self,v):
        """Calculate the permittivity at a given frequency.

        Parameters
        ----------
        v : float
            The frequency in internal units.

        Returns
        -------
        ndarray
            The permittivity at frequency v as a 3x3 tensor.

        """
        return self.value * np.eye(3) + self.epsilon_infinity

class ConstantScalar(DielectricFunction):
    """A complex constant scalar permittivity.

    Inherits from DielectricFunction. Provides a calculate() function to return
    the permittivity.  Epsilon infinity is not used for this function.

    Methods
    -------
    calculate()
        Calculates and returns the permittivity of the dielectric function.

    """

    def __init__(self, value):
        """Intialise ConstantScalar dielectric function.

        Parameters
        ----------
        value : float
            Initialise the permittivity with this tensor

        """
        DielectricFunction.__init__(self)
        self.value = value
        self.isScalarFunction = True

    def calculate(self,v):
        """Calculate the permittivity at a specified frequency.

        Parameters
        ----------
        v : float
            The frequency at which to calculate the permittivity.

        Returns
        -------
        float
            The permittivity at frequency v.

        """
        return self.value

class TabulateScalar(DielectricFunction):
    """A complex tabulated scalar permittivity.

    Inherits from DielectricFunction. 
    Represents an isotropic material dielectric function. 
    Provides a calculate() function to return the permittivity.
    Epsilon infinity is not used in the calculation.

    Methods
    -------
    calculate()
        Compute the permittivity of the material.

    Notes
    -----
    None

    Examples
    --------
    None

    """

    def __init__(self, vs_cm1, permittivities):
        """Intialise TabulateScalar dielectric function.

        Parameters
        ----------
        vs_cm1 : list of float
            Frequencies in cm-1 of the following permittivities.
        permittivities : list of complex
            The permittivities at each of the frequencies.

        """
        DielectricFunction.__init__(self)
        self.isScalarFunction = True
        self.vs_cm1 = np.array(vs_cm1)
        eps = np.array(permittivities)
        eps_r = np.real(eps)
        eps_i = np.imag(eps)
        self.interpolater = interpolate.CubicSpline(self.vs_cm1,eps_r)
        self.interpolatei = interpolate.CubicSpline(self.vs_cm1,eps_i)

    def calculate(self,v):
        """Calculate the permittivity at a given frequency.

        Parameters
        ----------
        v : float
            The frequency in internal units.

        Returns
        -------
        float
            The permittivity at frequency v as a scalar.

        """
        return complex(self.interpolater(v),self.interpolatei(v))

class Tabulate1(DielectricFunction):
    """A complex tabulated isotropic tensor permittivity.

    Inherits from DielectricFunction.
    Represents an isotropic dielectric function.
    Provides a calculate() function to return the permittivity.
    Epsilon infinity is included in the total permittivity

    Methods
    -------
    calculate()
        Returns the permittivity of the dielectric function.

    """

    def __init__(self, vs_cm1, permittivities):
        """Intialise Tabulate1 dielectric function.

        Parameters
        ----------
        vs_cm1 : list of float
            Frequencies in cm-1 of the following permittivities.
        permittivities : list of complex
            The permittivities at each of the frequencies.

        """
        DielectricFunction.__init__(self)
        self.isScalarFunction = False
        self.vs_cm1 = np.array(vs_cm1)
        eps = np.array(permittivities)
        eps_r = np.real(eps)
        eps_i = np.imag(eps)
        self.interpolater = interpolate.CubicSpline(self.vs_cm1,eps_r)
        self.interpolatei = interpolate.CubicSpline(self.vs_cm1,eps_i)

    def calculate(self,v):
        """Calculate the isotropic permittivity at a given frequency.

        Parameters
        ----------
        v : float
            The frequency in internal units.

        Returns
        -------
        numpy.ndarray
            The isotropic permittivity at frequency v as a 3x3 tensor.

        """
        eps = np.zeros( (3,3), dtype=complex )
        eps[0,0] = complex(self.interpolater(v),self.interpolatei(v))
        eps[1,1] = eps[0,0]
        eps[2,2] = eps[0,0]
        return eps + self.epsilon_infinity

class Tabulate3(DielectricFunction):
    """A complex tabulated diagonal tensor permittivity.

    Inherits from DielectricFunction.
    Represents an orthorhombic dielectric function.
    Provides a calculate() function to return the permittivity.
    Epsilon infinity is included in the total permittivity

    Methods
    -------
    calculate()
        Return the permittivity of the dielectric function.

    """

    def __init__(self, vs_cm1, epsxx, epsyy, epszz):
        """Intialise a Tabulate3 dielectric function object.

        Parameters
        ----------
        vs_cm1 : list of floats
            frequencies in cm-1 of the following permittivities
        epsxx : list of complex
            The eps(xx) permittivities at each of the frequencies
        epsyy : list of complex
            The eps(yy) permittivities at each of the frequencies
        epszz : list of complex
            The eps(zz) permittivities at each of the frequencies

        """
        DielectricFunction.__init__(self)
        self.isScalarFunction = False
        self.vs_cm1 = np.array(vs_cm1)
        epsd = [np.array(epsxx), np.array(epsyy), np.array(epszz)]
        self.interpolater = []
        self.interpolatei = []
        for eps in epsd:
            self.interpolater.append(interpolate.CubicSpline(self.vs_cm1,np.real(eps)))
            self.interpolatei.append(interpolate.CubicSpline(self.vs_cm1,np.imag(eps)))

    def calculate(self,v):
        """Calculate the permittivity at a given frequency.

        Parameters
        ----------
        v : float
            The frequency in internal units.

        Returns
        -------
        numpy.ndarray
            The permittivity at frequency v as a diagonal 3x3 tensor.

        """
        eps = np.zeros( (3,3), dtype=complex )
        eps[0,0] = complex(self.interpolater[0](v),self.interpolatei[0](v))
        eps[1,1] = complex(self.interpolater[1](v),self.interpolatei[1](v))
        eps[2,2] = complex(self.interpolater[2](v),self.interpolatei[2](v))
        return eps + self.epsilon_infinity

class Tabulate6(DielectricFunction):
    """A complex tabulated, anisotropic tensor permittivity.

    Inherits from DielectricFunction.
    Represents an non-isotropic dielectric function.
    Provides a calculate() function to return the permittivity.
    Epsilon infinity is included in the total permittivity

    Methods
    -------
    calculate()
        Returns the permittivity of the dielectric.

    """

    def __init__(self, vs_cm1, epsxx, epsyy, epszz, epsxy, epsxz, epsyz):
        """Intialise a Tabulate6 dielectric function object.

        Parameters
        ----------
        vs_cm1 : list of float
            The frequencies of the tabulated dielectric.
        epsxx : list of complex
            The eps(xx) permittivities at each of the frequencies.
        epsyy : list of complex
            The eps(yy) permittivities at each of the frequencies.
        epszz : list of complex
            The eps(zz) permittivities at each of the frequencies.
        epsxy : list of complex
            The eps(xy) permittivities at each of the frequencies.
        epsxz : list of complex
            The eps(xz) permittivities at each of the frequencies.
        epsyz : list of complex
            The eps(yz) permittivities at each of the frequencies.

        """
        DielectricFunction.__init__(self)
        self.isScalarFunction = False
        self.vs_cm1 = np.array(vs_cm1)
        eps6 = [np.array(epsxx), np.array(epsyy), np.array(epszz), np.array(epsxy), np.array(epsxz), np.array(epsyz)]
        self.interpolater = []
        self.interpolatei = []
        for eps in eps6:
            epsr = np.real(eps)
            epsi = np.imag(eps)
            self.interpolater.append(interpolate.CubicSpline(self.vs_cm1,epsr))
            self.interpolatei.append(interpolate.CubicSpline(self.vs_cm1,epsi))

    def calculate(self,v):
        """Calculate the symmetric permittivity tensor at a given frequency.

        Parameters
        ----------
        v : float
            The frequency in cm-1.

        Returns
        -------
        numpy.ndarray
            The full permittivity tensor at frequency v as a 3x3 tensor.

        """
        eps = np.zeros( (3,3), dtype=complex )
        eps[0,0] = complex(self.interpolater[0](v),self.interpolatei[0](v))
        eps[1,1] = complex(self.interpolater[1](v),self.interpolatei[1](v))
        eps[2,2] = complex(self.interpolater[2](v),self.interpolatei[2](v))
        eps[1,0] = complex(self.interpolater[3](v),self.interpolatei[3](v))
        eps[2,0] = complex(self.interpolater[4](v),self.interpolatei[4](v))
        eps[2,1] = complex(self.interpolater[5](v),self.interpolatei[5](v))
        eps[0,1] = eps[1,0]
        eps[0,2] = eps[2,0]
        eps[1,2] = eps[2,1]
        return eps + self.epsilon_infinity

class DFT(DielectricFunction):
    """A complex tensor permittivity from a DFT calculation.

    Inherits from DielectricFunction.
    Suitable for use in calculating the permittivity from DFT calculations.
    Internal units for this function are atomic units.
    Provides a calculate() function to return the permittivity.
    Epsilon infinity is included in the total permittivity

    Methods
    -------
    calculate()
        Returns the permittivity based on the DFT calculations.

    Notes
    -----
    The internal units for all calculations within this class are atomic units.

    """

    def __init__(self,mode_list, mode_frequencies, mode_sigmas, mode_oscillator_strengths, crystal_volume, drude, drude_plasma, drude_sigma):
        """Intialise a DFT dielectric function.

        Parameters
        ----------
        mode_list : list of int
            A list of active modes.
        mode_frequencies : list of float
            All mode frequencies.
        mode_sigmas : list of float
            All mode frequencies.
        mode_oscillator_strengths : list of float
            All mode oscillator strengths.
        crystal_volume : float
            The crystal volume.
        drude : bool
            True if a drude oscillator is to be included.
        drude_plasma : float
            Drude plasma frequency.
        drude_sigma : float
            Drude plasma width.
        All parameters are given in atomic units.

        """
        DielectricFunction.__init__(self)
        self.isScalarFunction = False
        self.mode_list = mode_list
        self.mode_frequencies = mode_frequencies
        self.mode_sigmas = mode_sigmas
        self.mode_oscillator_strengths = mode_oscillator_strengths
        self.crystal_volume = crystal_volume
        self.drude = drude
        self.drude_plasma = drude_plasma
        self.drude_sigma = drude_sigma

    def calculate(self,v):
        """Calculate the permittivity at a given frequency.

        Parameters
        ----------
        v : float
            The frequency in cm-1.

        Returns
        -------
        ndarray
            The permittivity at frequency v as a 3x3 tensor.

        """
        # For this class we force the use of atomic units and ignore internal units
        v_au = v * wavenumber
        eps = self.dielectriContributionsFromModes(v_au, self.mode_list, self.mode_frequencies, 
                                                   self.mode_sigmas, self.mode_oscillator_strengths, 
                                                   self.crystal_volume)
        if self.drude:
            eps = eps + self.dielectricContributionFromDrude(v_au, self.drude_plasma, self.drude_sigma, self.crystal_volume)
        return eps + self.epsilon_infinity

    def setFrequencies(self,frequencies):
        """Set the frequencies of a DFT permittivity.

        The frequencies are converted to atomic units

        Parameters
        ----------
        frequencies : 1d array of floats
            The frequencies in cm-1

        Returns
        -------
        None

        """
        frequencies = np.array(frequencies)
        self.mode_frequencies = wavenumber * frequencies
        return 


    def getFrequencies(self):
        """Return the frequencies of a DFT permittivity.

        The frequencies are in units of cm-1

        Parameters
        ----------
        None

        Returns
        -------
        1d array of floats
            The frequencies in cm-1

        """
        return self.mode_frequencies / wavenumber

    def setOscillatorStrengths(self,strengths):
        """Set the oscillator strengths of a DFT permittivity.

        The oscillator strengths are converted to atomic units

        Parameters
        ----------
        strengths : 3x3 array for each frequency
            The oscillator strengths parameters in cm-1

        Returns
        -------
        None

        """
        strengths = np.array(strengths)
        self.mode_oscillator_strengths = wavenumber * strengths
        return 


    def getOscillatorStrengths(self):
        """Return the oscillator strengths of a DFT permittivity.

        The returned strengths are in units of cm-1

        Parameters
        ----------
        None

        Returns
        -------
        3x3 array of floats for each frequency
            The oscillator strengths in cm-1

        """
        return self.mode_oscillator_strengths / wavenumber

    def setSigmas(self,sigmas):
        """Set the sigma parameters of a DFT permittivity.

        The sigmas parameter is converted to atomic units

        Parameters
        ----------
        sigmas : 1d array of floats
            The sigma parameters in cm-1

        Returns
        -------
        None

        """
        sigmas = np.array(sigmas)
        self.mode_sigmas = wavenumber * sigmas
        return 


    def getSigmas(self):
        """Return the sigma parameters of a DFT permittivity.

        The returned sigmas are in units of cm-1

        Parameters
        ----------
        None

        Returns
        -------
        1d array of floats
            The sigma parameters of  Drude-Lorentzian permittivity

        """
        return self.mode_sigmas / wavenumber


class DrudeLorentz(DielectricFunction):
    """A complex tensor Drude-Lorentz permittivity.

    Inherits from DielectricFunction.
    Provides a calculate() function to return the permittivity.
    Epsilon infinity is included in the total permittivity

    Methods
    -------
    calculate()
        Calculates and returns the permittivity.

    Examples
    --------
    None

    """

    def __init__(self, vs_cm1, strengths_cm1, sigmas_cm1):
        """Initialise a DrudeLorenz dielectric function.

        Parameters
        ----------
        vs_cm1 : list of ints
            Frequencies of the strengths and sigmas in cm-1.
        strengths_cm1 : list of floats
            strengths of isotropic contribution ([]).
            strengths of orthonormal contributions ([[xxs][yys][zzs]]), shape is either [] or [[xxs][yys][zzs]].
        sigmas_cm1 : list of floats
            Sigmas of isotropic contribution ([]).
            Sigmas of orthonormal contributions ([[xxs][yys][zzs]]), shape is either [] or [[xxs][yys][zzs]].

        """
        DielectricFunction.__init__(self)
        self.isScalarFunction = False
        rhombic = False
        if isinstance(vs_cm1[0], list):
            rhombic = True
        if rhombic:
            self.vs_cm1        = vs_cm1
            self.strengths_cm1 = strengths_cm1
            self.sigmas_cm1    = sigmas_cm1
        else:
            self.vs_cm1        = [vs_cm1, vs_cm1, vs_cm1]
            self.strengths_cm1 = [strengths_cm1, strengths_cm1, strengths_cm1]
            self.sigmas_cm1    = [sigmas_cm1, sigmas_cm1, sigmas_cm1]
        return

    def calculate(self, f_cm1):
        """Calculate the permittivity at a given frequency.

        Parameters
        ----------
        f_cm1 : float
            The frequency in cm-1

        Returns
        -------
        tensor
            The permittivity at frequency v as a diagonal 3x3 tensor.

        """
        eps = np.zeros( (3,3), dtype=complex )
        for xyz, (vs, strengths, sigmas) in enumerate(zip(self.vs_cm1,self.strengths_cm1, self.sigmas_cm1)):
            for v, strength, sigma in zip(vs, strengths,sigmas):
                eps[xyz,xyz] += strength * strength / complex((v*v - f_cm1*f_cm1), -sigma*f_cm1)
        return eps + self.epsilon_infinity

    def setFrequencies(self,frequencies):
        """Set the frequencies of a Drude-Lorentz permittivity.

        The frequencies are in cm-1

        Parameters
        ----------
        frequencies : 1d array of floats
            The frequencies in cm-1

        Returns
        -------
        None

        """
        frequencies = np.array(frequencies)
        self.vs_cm1 = frequencies
        return 


    def getFrequencies(self):
        """Return the frequencies of a Drude-Lorentz permittivity.

        The frequencies are in units of cm-1

        Parameters
        ----------
        None

        Returns
        -------
        1d array of floats
            The frequencies in cm-1

        """
        return self.vs_cm1

    def setOscillatorStrengths(self,strengths):
        """Set the oscillator strengths of a Drude-Lorentz permittivity.

        The oscillator strengths are in cm-1

        Parameters
        ----------
        strengths : 1d array of floats
            The oscillator strengths parameters in cm-1

        Returns
        -------
        None

        """
        strengths = np.array(strengths)
        self.strengths = strengths
        return 

    def getOscillatorStrengths(self):
        """Return the oscillator strengths of a Drude-Lorentz permittivity.

        The returned strengths are in units of cm-1

        Parameters
        ----------
        None

        Returns
        -------
        1d array of floats
            The oscillator strengths in cm-1

        """
        return self.strengths

    def setSigmas(self,sigmas):
        """Set the sigma parameters of a Drude-Lorentz permittivity.

        The default is to print an error message unless the DielectricFunction is a Lorentzian

        Parameters
        ----------
        sigmas : 1d array of floats
            The sigma parameters in cm-1

        Returns
        -------
        None

        """
        self.sigmas_cm1 = sigmas
        return 


    def getSigmas(self):
        """Return the sigma parameters of a Lorentzian permittivity.

        The returned sigmas are in units of cm-1

        Parameters
        ----------
        None

        Returns
        -------
        1d array of floats

        """
        return self.sigmas_cm1


class FPSQ(DielectricFunction):
    """A complex tensor FPSQ dielectric function.

    Inherits from DielectricFunction.
    Provides a calculate() function to return the permittivity.
    Epsilon infinity is included in the total permittivity
    """

    def __init__(self, omega_tos, gamma_tos, omega_los, gamma_los):
        """Initialise an FPS dielectric function object.

        Parameters
        ----------
        omega_tos : list of floats
            Omega (to) in the FPSQ equation.
        gamma_tos : list of floats
            Gamma (to) in the FPSQ equation.
        omega_los : list of floats
            Omega (lo) in the FPSQ equation.
        gamma_los : list of floats
            Gamma (lo) in the FPSQ equation.

        """
        DielectricFunction.__init__(self)
        self.isScalarFunction = False
        rhombic = False
        if isinstance(omega_tos[0], list):
            rhombic = True
        if rhombic:
            self.omega_tos = omega_tos
            self.gamma_tos = gamma_tos
            self.omega_los = omega_los
            self.gamma_los = gamma_los
        else:
            self.omega_tos = [omega_tos, omega_tos, omega_tos]
            self.gamma_tos = [gamma_tos, gamma_tos, gamma_tos]
            self.omega_los = [omega_los, omega_los, omega_los]
            self.gamma_los = [gamma_los, gamma_los, gamma_los]
        return

    def calculate(self, v):
        """Calculate the diagonal permittivity 3x3 tensor at a given frequency.

        Parameters
        ----------
        v : float
            The frequency in cm-1

        Returns
        -------
        numpy.ndarray
            The diagonal permittivity 3x3 tensor at frequency v.

        """
        f_cm1 = v
        eps = np.array(self.epsilon_infinity,dtype=complex)
        for xyz,(omega_tos,gamma_tos,omega_los,gamma_los) in enumerate(zip(self.omega_tos,self.gamma_tos,self.omega_los, self.gamma_los)):
            for omega_to, gamma_to, omega_lo, gamma_lo  in zip(omega_tos,gamma_tos,omega_los,gamma_los):
                contribution = (omega_lo**2 - f_cm1**2 - complex(0,gamma_lo)*f_cm1)/(omega_to**2 - f_cm1**2 - complex(0,gamma_to)*f_cm1)
                # make sure the imaginary component is not negative!
                real_contribution = np.real(contribution)
                imag_contribution = max(-1.0E-12,np.imag(contribution))
                eps[xyz,xyz] *= complex(real_contribution,imag_contribution)
        return eps 

class Sellmeier(DielectricFunction):
    """A real tensor Sellmeier permittivity.

    Inherits from DielectricFunction. Provides a calculate() function to return the permittivity.
    Epsilon infinity is not included in the total permittivity

    """

    def __init__(self, Bs, Cs):
        """Initialise an Sellmeier dielectric function object.

        Parameters
        ----------
        Bs : list of floats
            B parameter for Sellmeier equation (microns^2)
        Cs : list of floats
            C parameter for Sellmeier equation (microns^2)

        """
        DielectricFunction.__init__(self)
        self.isScalarFunction = True
        self.Bs = Bs
        self.Cs = Cs
        return

    def calculate(self, v):
        """Calculate the diagonal permittivity 3x3 tensor at a given frequency.

        Parameters
        ----------
        v : float
            The frequency in cm-1.

        Returns
        -------
        ndarray
            The diagonal permittivity 3x3 tensor at frequency v.

        Notes
        -----
        The Sellmeier parameters are in microns^2.

        """
        f_cm1 = v
        # Convert to wavelength in microns
        wavelength_mu = f_cm1 * 1.0E+6 *1.0E-2            # convert cm-1 to microns
        wavelength2 = wavelength_mu*wavelength_mu
        n2 = 1.0
        for B,C in zip(self.Bs,self.Cs):
            n2 += B*wavelength2/(wavelength2-C) 
        return n2 




#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""PlottingTab module."""

# Import plotting requirements
import matplotlib
import matplotlib.figure
import numpy as np
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar
from qtpy.QtCore import QCoreApplication, Qt
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QDoubleSpinBox,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from PDielec.Constants import avogadro_si
from PDielec.Utilities import Debug

possible_frequency_units = ["wavenumber","THz","GHz","ang","nm","um","mm","cm","m"]

def isThisAFrequencyUnit(unit):
    """Return true if this is a frequency unit, false if a wavelength.

    Units of frequency are 'wavenumber','THz','GHz'
    Units of wavelength are 'ang','nm','um','mm','cm' or 'm'

    Returns
    -------
    bool
        True if this is a frequency unit, False if a wavelength.

    """
    index = possible_frequency_units.index(unit)
    return index <= 2

def convert_frequency_units( value, unit_in, unit_out ):
    """Convert between frequency and wavelength units.

    The input can be either a scalar value or a numpy array of values. The function will return the converted value(s) in the output units specified.
    The unit strings are turned into lower-case so case is irrelevant

    Parameters
    ----------
    value : scalar or numpy array
        The value(s) for which the conversion is to be made.
    unit_in : str
        The units of the input value(s). Can be one of 'cm-1' (or 'wavenumber'), 'GHz', 'THz', 'nm', 'um', 'mm', 'cm', 'm'.
    unit_out : str
        The units of the output value(s). Must be one of 'cm-1' (or 'wavenumber'), 'GHz', 'THz', 'nm', 'um', 'mm', 'cm', 'm'.

    Returns
    -------
    scalar or numpy array
        The converted value(s) in the output units specified.

    """
    unit_in  = unit_in.lower()
    unit_out = unit_out.lower()
    # the conversion dictionary has a tuple for each unit
    # the first entry is the multiplicative factor, the second is the operator
    # To convert wavelength they have to be scaled and then inverted
    wavenumber = { "cm-1"       : (1,"*"),
                   "wavenumber" : (1,"*"),
                   "thz"        : (33.356,"*"),
                   "ghz"        : (33.356E+3,"*"),
                   "ang"        : (1E-8,"/"),
                   "nm"         : (1E-7,"/"),
                   "um"         : (1E-4,"/"),
                   "mm"         : (1E-1,"/"),
                   "cm"         : (1.0 ,"/"),
                   "m"          : (1E2 ,"/"),
                 }
    # Convert the input unit to cm-1
    if isinstance(value,np.ndarray):
        for i,x in enumerate(value):
            if x <= 0:
                # Problem with negative or zero conversions between wavelength and frequency
                print("Warning a zero or negative frequency/wavelength is not permitted",i,x,unit_in,unit_out)
                x[i] = 1.0E-8
    elif value <= 0 :
        value = 1.0E-8
    scale,operator = wavenumber[unit_in]
    value = scale * value     
    if operator == "/":
        value = 1.0 / value  
    # convert the internal value from cm-1 to the output unit
    scale,operator = wavenumber[unit_out]
    if operator == "/":
        value = 1.0 / value  
    return value / scale
    
class PlottingTab(QWidget):
    """A class used for creating and managing a plotting tab in a graphical user interface. It inherits from QWidget.

    This class is responsible for handling plotting functionalities like choosing plot types, setting molar definitions and frequency values, managing interactions with UI components such as spin boxes, combo boxes, and buttons, and plotting the data using matplotlib.

    Parameters
    ----------
    parent : QWidget
        The parent widget.
    debug : bool, optional
        Enables debug mode which provides additional logging details. Defaults to False.

    Attributes
    ----------
    settings : dict
        A dictionary containing settings related to plotting such as minimum frequency, maximum frequency, frequency increment, molar definitions, number of atoms, plot type, and frequency unit.
    refreshRequired : bool
        A flag indicating whether the plot needs to be refreshed.
    subplot : matplotlib subplot object
        The subplot used for plotting data.
    vmin, vmax, vinc : float
        Variables to store current minimum frequency, maximum frequency, and frequency increment values for the plot.
    molar_cb_current_index : int
        Tracks the current index of the molar definition combo box to handle changes.
    notebook : QWidget
        A reference to the parent widget which contains the widget structure.
    reader : object
        An object to read and handle data from different file formats or sources.
    legends, vs_cm1 : list
        Lists used for storing legends for the plots and frequency values respectively.
    frequency_length : int
        Stores the length of the frequency array to check for changes.
    progressbar : QProgressBar
        A progress bar widget to display operation progress.

    Methods
    -------
    get_total_number_of_frequency_calculations()
        Calculates the total number of frequency calculations required.
    requestRefresh()
        Requests a refresh for the plotting.
    requestScenarioRefresh()
        Requests a refresh for all scenarios in the application.
    on_vmin_changed(), on_vmax_changed(), on_vinc_changed(value), on_funits_cb_activated(index), on_molar_cb_activated(index), on_natoms_changed(value), on_plot_type_cb_activated(index)
        Event handlers for UI component changes.
    refresh(force=False)
        Refreshes the plot based on current settings and data.
    plot()
        Generates and displays the plot based on the selected plot type and data.
    get_number_of_califications_required()
        Returns the number of calculations needed for updating the plot.
    greyed_out()
        Greys out options in the UI that are not available.
    writeSpreadsheet()
        Writes results to a spreadsheet.
    set_concentrations()
        Sets the concentrations for the plotting based on the molar definition.

    """

    def __init__(self, parent, debug=False ):
        """Initialise the Plotting tab within a given parent.

        Parameters
        ----------
        parent : QWidget
            The parent widget in which this plotting tab will be initialized.
        debug : bool, optional
            Determines whether debugging messages should be shown, by default False.

        Attributes
        ----------
        settings : dict
            A dictionary to store settings such as frequency range, molar definitions, etc.
        refreshRequired : bool
            Indicates whether the plot needs to be refreshed.
        subplot : NoneType or matplotlib subplot
            Placeholder for a future subplot object, initialized as None.
        molar_definitions : list
            A list of possible definitions for a mole, including 'Unit cells', 'Atoms', 'Molecules'.
        legends : list
            A list to store legend entries.
        vs_cm1 : list
            A list to store frequency data.
        frequency_length : int
            Stores the length of `vs_cm1` list.
        vmin, vmax, vinc : float
            Variables to store minimum, maximum, and increment values for frequency/wavelength after conversion.
        molar_cb_current_index : int
            Tracks the currently selected index in the molar definitions combo box.
        notebook : QWidget
            A reference to the parent widget, here it's assumed to be a notebook containing multiple tabs.
        reader : Object
            An object that deals with reading data, accessed through the `notebook`'s mainTab.

        Notes
        -----
        Widget elements such as QDoubleSpinBox, QComboBox, QLabel, QPushButton, QProgressBar, and the matplotlib figure are also initialized and configured within this method. This involves setting up signal-slot connections for interactive behavior, tooltips for user guidance, and incorporating the matplotlib canvas and toolbar for plotting functionality. Additionally, the method is designed for setup within a Qt layout structure, ensuring proper arrangement and display of the interface components.

        """        
        super(QWidget, self).__init__(parent)
        global debugger
        debugger = Debug(debug,"PlottingTab")
        debugger.print("Start:: Plotting tab initialisation")
        self.settings = {}
        self.refreshRequired = True
        self.subplot = None
        self.setWindowTitle("Plotting")
        self.settings["Minimum frequency"] = 1.0
        self.settings["Maximum frequency"] = 200
        self.settings["Frequency increment"] = 0.2
        self.molar_definitions = ["Unit cells","Atoms","Molecules"]
        self.settings["Molar definition"] = "Unit cells"
        self.settings["Number of atoms"] = 1
        self.settings["Plot type"] = "Powder Molar Absorption"
        self.settings["Frequency unit"] = "wavenumber"
        # self.settings['Plot title'] = 'Plot Title'
        self.legends = []
        self.vs_cm1 = []
        self.frequency_length = len(self.vs_cm1)
        self.vmin = 0.0
        self.vmax = 0.0
        self.vinc = 0.0
        self.molar_cb_current_index = 0
        # store the notebook
        self.notebook = parent
        # get the reader from the main tab
        self.reader = self.notebook.mainTab.reader
        # Create last tab - PlottingTab
        vbox = QVBoxLayout()
        form = QFormLayout()
        # Preamble for frequencies / wavelengths display
        vmin = self.settings["Minimum frequency"]
        vmax = self.settings["Maximum frequency"]
        vinc = self.settings["Frequency increment"]
        # n is the number of samples
        n = len(np.arange(float(vmin), float(vmax)+0.5*float(vinc), float(vinc)))
        # convert units
        vmin = convert_frequency_units(vmin,"wavenumber",self.settings["Frequency unit"])
        vmax = convert_frequency_units(vmax,"wavenumber",self.settings["Frequency unit"])
        # vinc is calculated using the number of samples
        vinc = (vmax-vmin) / n
        vinc = convert_frequency_units(vinc,"wavenumber",self.settings["Frequency unit"])
        # If dealing with wavelength then swap the order of vmin and vmax
        if not isThisAFrequencyUnit(self.settings["Frequency unit"]):
            vmin, vmax = vmax, vmin
        #
        # The minimum frequency
        #
        self.vmin_sb = QDoubleSpinBox(self)
        self.vmin_sb.setRange(0.00000001,900000000000)
        self.vmin_sb.setValue(vmin)
        if isThisAFrequencyUnit(self.settings["Frequency unit"]):
            self.vmin_sb.setToolTip("Set the minimum frequency to be considered)")
        else:
            self.vmin_sb.setToolTip("Set the minimum wavelength to be considered)")
        self.vmin_sb.valueChanged.connect(self.on_vmin_changed)
        #
        # The maximum frequency
        #
        self.vmax_sb = QDoubleSpinBox(self)
        self.vmax_sb.setRange(0.00000001,900000000000)
        self.vmax_sb.setValue(vmax)
        if isThisAFrequencyUnit(self.settings["Frequency unit"]):
            self.vmax_sb.setToolTip("Set the maximum frequency to be considered)")
        else:
            self.vmax_sb.setToolTip("Set the maximum wavelength to be considered)")
        self.vmax_sb.valueChanged.connect(self.on_vmax_changed)
        #
        # Choose a suitable increment
        #
        self.vinc_sb = QDoubleSpinBox(self)
        self.vinc_sb.setRange(0.0000001,5000000000.0)
        self.vinc_sb.setSingleStep(0.1)
        self.vinc_sb.setDecimals(4)
        if isThisAFrequencyUnit(self.settings["Frequency unit"]):
            self.vinc_sb.setToolTip("Choose an increment for the frequency when plotting")
        else:
            self.vinc_sb.setToolTip("Choose an increment for the wavelength when plotting")
        self.vinc_sb.setValue(vinc)
        self.vinc_sb.valueChanged.connect(self.on_vinc_changed)
        #
        # Set the frequency units
        #
        self.funits_cb = QComboBox(self)
        self.funits_cb.addItems( possible_frequency_units )
        self.funits_cb.activated.connect(self.on_funits_cb_activated)
        index = possible_frequency_units.index(self.settings["Frequency unit"])
        self.funits_cb.setCurrentIndex(index)
        if isThisAFrequencyUnit(self.settings["Frequency unit"]):
            self.funits_cb.setToolTip("Set the frequency unit")
            self.frequency_form_label = QLabel("Frequency min, max and increment", self)
            self.frequency_form_label.setToolTip("Choose minimum, maximum and increment for frequency")
        else:
            self.funits_cb.setToolTip("Set the wavelength unit")
            self.frequency_form_label = QLabel("Wavelength min, max and increment", self)
            self.frequency_form_label.setToolTip("Choose minimum, maximum and increment for wavelength")
        hbox = QHBoxLayout()
        hbox.addWidget(self.vmin_sb)
        hbox.addWidget(self.vmax_sb)
        hbox.addWidget(self.vinc_sb)
        hbox.addWidget(self.funits_cb)
        form.addRow(self.frequency_form_label, hbox)
        #
        # Define molar quantity
        #
        self.molar_cb = QComboBox(self)
        self.molar_cb.setToolTip("Define what a mole is.  \nIn the case of Molecules, the number of atoms in a molecule must be given")
        self.molar_cb.addItems(self.molar_definitions)
        try:
            self.molar_cb_current_index = self.molar_definitions.index(self.settings["Molar definition"])
        except Exception:
            self.molar_cb_current_index = 0
            self.settings["Molar definition"] = self.molar_definitions[self.molar_cb_current_index]
        self.molar_cb.setCurrentIndex(self.molar_cb_current_index)
        self.molar_cb.activated.connect(self.on_molar_cb_activated)
        label = QLabel("Molar definition", self)
        label.setToolTip("Define what a mole is.  \nIn the case of Molecules, the number of atoms in a molecule must be given")
        form.addRow(label, self.molar_cb)
        #
        # Number of atoms in a molecule
        #
        self.natoms_sb = QSpinBox(self)
        self.natoms_sb.setToolTip("Set the number of atoms in a molecule. \nOnly need this if moles of molecules is needed")
        self.natoms_sb.setRange(1,500)
        self.natoms_sb.setValue(self.settings["Number of atoms"])
        self.natoms_sb.valueChanged.connect(self.on_natoms_changed)
        self.natoms_sb.setEnabled(False)
        label = QLabel("Number of atoms per molecule", self)
        label.setToolTip("Set the number of atoms in a molecule. \nOnly need this if moles of molecules is needed")
        form.addRow(label, self.natoms_sb)
        #
        # Final button
        #
        self.plot_type_cb = QComboBox(self)
        self.plot_type_cb.setToolTip("Choose the which data to plot")
        self.plot_types = [
                            "Powder Molar Absorption",
                            "Powder Absorption",
                            "Powder Real Permittivity",
                            "Powder Imaginary Permittivity",
                            "Powder ATR",
                            "Crystal Reflectance (P polarisation)",
                            "Crystal Reflectance (S polarisation)",
                            "Crystal Transmittance (P polarisation)",
                            "Crystal Transmittance (S polarisation)",
                            "Crystal Absorbtance (P polarisation)",
                            "Crystal Absorbtance (S polarisation)",
                          ]
        self.plot_ylabels = {
                     "Powder Molar Absorption": r"Molar Absorption Coefficient $\mathdefault{(L mole^{-1} cm^{-1})}$",
                           "Powder Absorption": r"Absorption Coefficient $\mathdefault{(cm^{-1})}$",
                    "Powder Real Permittivity": r"Real Component of Permittivity",
               "Powder Imaginary Permittivity": r"Imaginary Component of Permittivity",
                                  "Powder ATR": r"ATR absorption",
        "Crystal Reflectance (P polarisation)": r"Fraction of p-polarised reflectance",
        "Crystal Reflectance (S polarisation)": r"Fraction of s-polarised reflectance",
      "Crystal Transmittance (P polarisation)": r"Fraction of p-polarised transmitted",
      "Crystal Transmittance (S polarisation)": r"Fraction of s-polarised transmitted",
        "Crystal Absorbtance (P polarisation)": r"Fraction of p-polarised absorbtance",
        "Crystal Absorbtance (S polarisation)": r"Fraction of s-polarised absorbtance",
                            }

        self.plot_type_cb.activated.connect(self.on_plot_type_cb_activated)
        self.plot_type_cb.addItems( self.plot_types )
        label = QLabel("Choose plot type", self)
        label.setToolTip("Choose the plot type")
        index = self.plot_type_cb.findText(self.settings["Plot type"], Qt.MatchFixedString)
        self.plot_type_cb.setCurrentIndex(index)
        plot_button = QPushButton("Update plot")
        plot_button.clicked.connect(self.refresh)
        plot_button.setToolTip("Update the plot")
        hbox = QHBoxLayout()
        hbox.addWidget(self.plot_type_cb)
        hbox.addWidget(plot_button)
        form.addRow(label, hbox)
        # Add a progress bar
        self.progressbar = QProgressBar(self)
        self.progressbar.setToolTip("Show the progress of any calculations")
        # Append the progress bar to the list of progress bars managed by the notebook
        self.notebook.progressbars_add(self.progressbar)
        self.notebook.progressbars_set_maximum(self.get_total_number_of_frequency_calculations())
        label = QLabel("Calculation progress", self)
        label.setToolTip("Show the progress of any calculations")
        form.addRow(label,self.progressbar)
        # Add the matplotlib figure to the bottom
        self.figure = matplotlib.figure.Figure()
        self.canvas = FigureCanvas(self.figure)
        self.canvas.setSizePolicy(QSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding))
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.toolbar.setSizePolicy(QSizePolicy(QSizePolicy.Fixed,QSizePolicy.Fixed))
        form.addRow(self.canvas)
        form.addRow(self.toolbar)
        vbox.addLayout(form)
        # finalise the layout
        self.setLayout(vbox)
        QCoreApplication.processEvents()
        # Create the plot
        debugger.print("Finished:: Plotting tab initialisation")
        return

    def get_total_number_of_frequency_calculations(self):
        """Calculate and return the total number of frequency calculations required.

        This method computes the total number of frequency calculations based on the current settings for minimum frequency, maximum frequency, and frequency increment. It updates the scenario if the frequency range or the increment has changed since the last update.

        Parameters
        ----------
        None

        Returns
        -------
        int
            The total number of frequency calculations required, which is the product of the number of calculations required (as obtained from `get_number_of_calculations_required`) and the new length of frequency range computed.

        """        
        vmin = self.settings["Minimum frequency"]
        vmax = self.settings["Maximum frequency"]
        vinc = self.settings["Frequency increment"]
        self.vs_cm1 = np.arange(float(vmin), float(vmax)+0.5*float(vinc), float(vinc))
        new_length = len(self.vs_cm1)
        if self.frequency_length != new_length or self.vmin != vmin or self.vmax != vmax or self.vinc != vinc :
            self.requestScenarioRefresh()
            self.frequency_length = new_length
            self.vmin = vmin
            self.vmax = vmax
            self.vinc = vinc
        n = self.get_number_of_calculations_required()
        return n*new_length

    def requestRefresh(self):
        """Initiate a refresh request.

        This function flags that a refresh is required
        It doesn't return any value.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """        
        debugger.print("Start:: requestRefresh")
        self.refreshRequired = True
        debugger.print("Finished:: requestRefresh")
        return

    def requestScenarioRefresh(self):
        """Request a refresh on all scenarios within a notebook.

        This function triggers a refresh process for the settings tab and all scenarios within the notebook. 

        Parameters
        ----------
        None

        Returns
        -------
        None

        """        
        debugger.print("Start:: requestScenarioRefresh")
        self.notebook.settingsTab.requestRefresh()
        for scenario in self.notebook.scenarios:
            scenario.requestRefresh()
        debugger.print("Finished:: requestScenarioRefresh")
        return

    def on_vinc_changed(self,value):
        """Handle the change in frequency increment and update GUI accordingly.

        This function is triggered when there's a change in the frequency increment value. It adjusts the number of GUI elements based on 
        new values of minimum and maximum frequency and the changed frequency increment. It also updates the settings dict with the 
        new frequency increment value.

        Parameters
        ----------
        value : float
            The new value of the frequency increment.

        Returns
        -------
        None

        Side Effects
        ------------
        - Temporarily blocks signals from the vinc_sb spin box to prevent recursive calls.
        - Updates 'Frequency increment' in the settings dictionary with the new frequency increment value.
        - Requests a refresh of the fitterTab if needed.

        """        
        debugger.print("Start:: on_vinc_changed", value)
        if value <=0 :
            return
        self.vinc_sb.blockSignals(True)
        # Use existing units to calculate the number of samples
        vmin = self.vmin_sb.value()
        vmax = self.vmax_sb.value()
        ngui = int((vmax - vmin) / value) + 1
        # Work out the increment needed for cm-1
        vmin = self.settings["Minimum frequency"]
        vmax = self.settings["Maximum frequency"]
        if ngui <= 2:
            return
        vinc = (vmax - vmin) / (ngui - 1)
        self.settings["Frequency increment"] = vinc
        self.notebook.fitterTab.requestRefresh()
        self.refreshRequired = True
        debugger.print("on_vinc_change ", self.settings["Frequency increment"])
        self.vinc_sb.blockSignals(False)
        debugger.print("Finished:: on_vinc_changed", value)

    def on_vmin_changed(self):
        """Handle the change in the minimum frequency setting.

        This method adjusts the stored minimum frequency in the settings according to the new
        minimum frequency input by the user. It first blocks the signals of the minimum frequency spinbox
        to prevent recursive calls during the adjustment process. It then updates the setting based
        on whether the frequency units are compatible with a predefined list, converting the frequency
        accordingly. After adjusting the settings, it triggers a refresh request for the application's
        fitter tab and marks the need for a general refresh. Finally, it re-enables the signals for
        the minimum frequency spinbox and logs the completion of the process.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        - Uses the `blockSignals` method on `self.vmin_sb` to prevent signal-slot recursion.
        - Adjusts settings based on a unit conversion utility function, `convert_frequency_units`.
        - The effect of this method extends beyond just the internal state changes; it influences the UI and possibly other components' states through the requested refresh.

        """        
        debugger.print("Start:: on_vmin_changed")
        self.vmin_sb.blockSignals(True)
        vmin = self.vmin_sb.value()
        if isThisAFrequencyUnit(self.settings["Frequency unit"]):
            self.settings["Minimum frequency"] = convert_frequency_units(vmin,self.settings["Frequency unit"],"wavenumber")
        else:
            self.settings["Maximum frequency"] = convert_frequency_units(vmin,self.settings["Frequency unit"],"wavenumber")
        debugger.print("on_vmin_changed setting vmin to", self.settings["Minimum frequency"])
        self.notebook.fitterTab.requestRefresh()
        self.refreshRequired = True
        self.vmin_sb.blockSignals(False)
        debugger.print("Finished:: on_vmin_changed")

    def on_vmax_changed(self):
        """Handle the event when the maximum frequency setting is changed.

        This method is triggered whenever there is a change in the maximum frequency setting (`vmax`). It blocks signal emission from the `vmax_sb` spinner box (presumably, a GUI element for setting `vmax`), reads the current minimum (`vmin`) and maximum (`vmax`) frequency settings, and updates the corresponding setting based on the specified frequency unit. If the frequency unit denotes a frequency, `vmax` is converted and stored as the 'Maximum frequency' in the settings dictionary in the 'wavenumber' unit. Otherwise, it updates the 'Minimum frequency' with the converted value. This method also triggers any necessary refresh operations in the user interface and finally unblocks signal emissions from the `vmax_sb`.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        - This method assumes the existence of `vmin_sb` and `vmax_sb` attributes, which should be spinner box GUI elements (or similar) for setting minimum and maximum frequencies, respectively.
        - The `settings` dictionary must have a 'Frequency unit' key, and possibly 'Maximum frequency' and 'Minimum frequency' keys which are updated based on the condition.

        """        
        debugger.print("Start:: on_vmax_changed")
        self.vmax_sb.blockSignals(True)
        vmax = self.vmax_sb.value()
        if isThisAFrequencyUnit(self.settings["Frequency unit"]):
            self.settings["Maximum frequency"] = convert_frequency_units(vmax,self.settings["Frequency unit"],"wavenumber")
        else:
            self.settings["Minimum frequency"] = convert_frequency_units(vmax,self.settings["Frequency unit"],"wavenumber")
        debugger.print("on_vmax_changed setting vmax to ", self.settings["Maximum frequency"])
        self.notebook.fitterTab.requestRefresh()
        self.refreshRequired = True
        self.vmax_sb.blockSignals(False)
        debugger.print("Finished:: on_vmax_changed")

    def refresh(self,force=False):
        """Refresh the current state based on the changes in settings, forces refresh if needed.

        This function updates the frequency settings (minimum frequency, maximum frequency, and frequency increment) based on the current settings. It ensures that these settings are within logical limits. The function updates GUI elements (spin boxes and combo boxes) with these new calculations. Additionally, it handles the conversion of frequency units, updates GUI tooltips based on the frequency unit, and builds or refreshes the visualization plot based on the refreshed data. If the 'force' flag is set or if a refresh is deemed necessary due to significant changes in the settings, it proceeds with the refresh operation, else it skips the refresh to optimize performance.

        Parameters
        ----------
        force : bool, optional
            A boolean flag indicating whether to forcefully execute a refresh regardless of whether it was deemed necessary based on internal conditions. The default is False.

        Returns
        -------
        None

        Notes
        -----
        - The actual refresh operation involves several steps:
            - Checking if a refresh is required based on the current settings and the 'force' parameter.
            - Temporarily blocking signals from all child widgets to prevent unintended side effects during settings updates.
            - Verifying and updating frequency settings to maintain logical constraints.
            - Updating GUI components with new settings values and tooltips based on the current frequency unit.
            - Processing and plotting data based on the updated settings.
            - Re-enabling signals for child widgets after modifications are complete.
        - The processEvents call is used to ensure the UI remains responsive during long operations.

        """        
        debugger.print("Start:: refresh", force)
        if not self.refreshRequired and not force:
            self.plot()
            debugger.print("Finished:: refreshing widget not required")
            return
        #
        # Block signals during refresh
        #
        self.greyed_out()
        for w in self.findChildren(QWidget):
            w.blockSignals(True)
        # Now refresh values
        if self.settings["Maximum frequency"] < self.settings["Minimum frequency"]:
            self.settings["Maximum frequency"] = self.settings["Minimum frequency"]+1
        if self.settings["Frequency increment"] > self.settings["Maximum frequency"] - self.settings["Minimum frequency"]:
            self.settings["Frequency increment"] = (self.settings["Maximum frequency"] - self.settings["Minimum frequency"])/2
        # The GUI uses frequency and wavelength, convert the values in settings from cm-1
        # Work out the increment needed for cm-1
        vmin = self.settings["Minimum frequency"]
        vmax = self.settings["Maximum frequency"]
        vinc = self.settings["Frequency increment"]
        # Protect the code from over-exuberant choice of parameters
        if (vmax - vmin)/vinc > 90000:
            vinc = (vmax - vmin) / 90000
            print("Warning - the number data points in a plot has been limited to 9000")
            print("          this happens if a 0 wavelength or frequency is entered in the GUI")
            self.settings["Frequency increment"] = vinc
        ncm1 = len(np.arange(float(vmin), float(vmax)+0.5*float(vinc), float(vinc)))
        vmin = convert_frequency_units(self.settings["Minimum frequency"],"wavenumber",self.settings["Frequency unit"])
        vmax = convert_frequency_units(self.settings["Maximum frequency"],"wavenumber",self.settings["Frequency unit"])
        # Reorder the GUI values
        if vmin > vmax:
            vmin, vmax = vmax, vmin
        # Set the GUI values
        vinc = (vmax - vmin) / (ncm1-1)
        self.vmin_sb.setValue(vmin)
        self.vmax_sb.setValue(vmax)
        self.vinc_sb.setValue(vinc)
        # Update the tool tips as the frequency unit could be a wavelength
        if isThisAFrequencyUnit(self.settings["Frequency unit"]):
            self.vmin_sb.setToolTip("Set the minimum frequency to be considered)")
            self.vmax_sb.setToolTip("Set the maximum frequency to be considered)")
            self.vinc_sb.setToolTip("Choose an increment for the frequency when plotting")
            self.funits_cb.setToolTip("Set the frequency unit")
            self.frequency_form_label.setToolTip("Choose minimum, maximum and increment for frequency")
            self.frequency_form_label.setText("Frequency min, max and increment")
            self.frequency_form_label.setToolTip("Choose minimum, maximum and increment for frequency")
        else:
            self.vmin_sb.setToolTip("Set the minimum wavelength to be considered)")
            self.vmax_sb.setToolTip("Set the maximum wavelength to be considered)")
            self.vinc_sb.setToolTip("Choose an increment for the wavelength when plotting")
            self.funits_cb.setToolTip("Set the wavelength unit")
            self.frequency_form_label.setText("Wavelength min, max and increment")
            self.frequency_form_label.setToolTip("Choose minimum, maximum and increment for wavelength")
        index = possible_frequency_units.index(self.settings["Frequency unit"])
        self.funits_cb.setCurrentIndex(index)
        index = self.plot_type_cb.findText(self.settings["Plot type"], Qt.MatchFixedString)
        self.plot_type_cb.setCurrentIndex(index)
        try:
            self.molar_cb_current_index = self.molar_definitions.index(self.settings["Molar definition"])
        except Exception:
            self.molar_cb_current_index = 0
            self.settings["Molar definition"] = self.molar_definitions[self.molar_cb_current_index]
        self.molar_cb.setCurrentIndex(self.molar_cb_current_index)
        self.natoms_sb.setValue(self.settings["Number of atoms"])
        # Refresh the widgets that depend on the reader
        self.reader = self.notebook.reader
        if self.reader is not None:
            self.set_concentrations()
        # Reset the progress bar
        self.notebook.progressbars_set_maximum(self.get_total_number_of_frequency_calculations())
        #
        # Unblock signals after refresh
        #
        for w in self.findChildren(QWidget):
            w.blockSignals(False)
        QCoreApplication.processEvents()
        debugger.print("calling plot from refresh")
        self.plot()
        self.refreshRequired = False
        debugger.print("Finished:: refresh", force)
        return

    def on_natoms_changed(self, value):
        """Handle the change in the number of atoms.

        This method is called when the number of atoms changes. It updates the relevant setting in the instance, recalculates the concentration based on the new number of atoms, flags that a refresh is required, and then triggers the refresh process.

        Parameters
        ----------
        value : int or float
            The new number of atoms. This value is used to update settings and recalculate concentrations.

        Notes
        -----
        - `self.reader.volume` and `self.reader.nions` are expected to be available and contain the volume of the container and the number of ions, respectively.
        - The method refreshes `self.notebook.fitterTab` object with a `requestRefresh` method to initiate the refresh process.

        """        
        debugger.print("Start:: on_natoms_changed", value)
        self.settings["Number of atoms"] = value
        debugger.print("on natoms changed ", self.settings["Number of atoms"])
        self.settings["concentration"] = 1000.0 / (avogadro_si * self.reader.volume * 1.0e-24 * self.settings["Number of atoms"] / self.reader.nions)
        debugger.print("The concentration has been set", self.settings["Molar definition"], self.settings["concentration"])
        self.refreshRequired = True
        self.notebook.fitterTab.requestRefresh()
        self.refresh()
        debugger.print("Finished:: on_natoms_changed", value)

    def on_plot_type_cb_activated(self, index):
        """Handle plot type change from a combo box.

        This method gets triggered when the plot type in a combo box is changed.
        It updates the plot type in the settings, marks the current state as needing a refresh,
        requests the fitter tab to refresh, and finally refreshes the current view.

        Parameters
        ----------
        index : int
            The index of the selected item in the combo box.

        Returns
        -------
        None

        """        
        debugger.print("Start:: on_plot_type_cb_activated", index)
        self.settings["Plot type"] = self.plot_type_cb.currentText()
        debugger.print("Changed plot type to ", self.settings["Plot type"])
        self.refreshRequired = True
        self.notebook.fitterTab.requestRefresh()
        self.refresh()
        debugger.print("Finished:: on_plot_type_cb_activated", index)

    def on_funits_cb_activated(self, index):
        """Handle the activation of a frequency unit combo box item.

        This method updates the active frequency unit in the settings, triggers necessary refresh processes and logs the change.

        Parameters
        ----------
        index : int
            The index of the activated item in the frequency units combo box.

        Returns
        -------
        None

        Notes
        -----
        This method is part of a GUI application where `self` refers to an instance of the application or a relevant widget. It manages updates to the application's settings and GUI components based on the user's selection of a frequency unit from a combo box.

        - `self.settings` is a dictionary where application settings are stored.
        - `self.refreshRequired` is a boolean flag used to indicate whether a refresh of certain GUI components is necessary.
        - `self.notebook` appears to be a widget container (like a tab widget), with `fitterTab` being one of its child tabs.
        - `self.vmin_sb` is another component (likely a spin box or similar input widget) that may need its signals blocked/unblocked during the process to avoid unwanted signal emission.
        - The `debugger.print` calls are used for logging and are not a standard Python function; they imply the existence of a custom logging or debugging utility named `debugger`.

        The actual refreshing of the GUI and handling of signal blocking is done within other methods not shown here, such as `self.refresh()` and `self.notebook.fitterTab.requestRefresh()`.

        Raises
        ------
        This function does not explicitly raise any exceptions but depends on the proper functioning of the methods it calls and the state of `self` and its attributes.

        """        
        debugger.print("Start:: on_funits_cb_activated", index)
        self.settings["Frequency unit"] = possible_frequency_units[index]
        self.refreshRequired = True
        self.notebook.fitterTab.requestRefresh()
        self.refreshRequired = True
        self.vmin_sb.blockSignals(False)
        self.refresh()
        debugger.print("Frequency unit changed to ", self.settings["Frequency unit"])
        debugger.print("Finished:: on_funits_cb_activated", index)

    def on_molar_cb_activated(self, index):
        """Handle the activation of the molar combobox option.

        This method is tied to a GUI event where a selection from a molar combobox triggers various updates in the application state, including setting concentrations, refreshing UI elements, and potentially triggering further calculations or refreshes as needed.

        Parameters
        ----------
        index : int
            The index of the selected item in the molar combobox. This index corresponds to a specific molar definition and is used to update the application settings and state accordingly.

        Returns
        -------
        None

        See Also
        --------
        set_concentrations : A method to update the concentrations based on the selected molar definition.
        refresh : A method to refresh the UI elements.

        """        
        debugger.print("Start:: on_molar_cb_activated", index)
        self.molar_cb_current_index = index
        self.settings["Molar definition"] = self.molar_definitions[index]
        self.set_concentrations()
        self.refreshRequired = True
        self.notebook.fitterTab.requestRefresh()
        self.refresh()
        debugger.print("The concentration has been set", self.settings["Molar definition"], self.settings["concentration"])
        debugger.print("Finished:: on_molar_cb_activated", index)
        return

    def set_concentrations(self):
        """Set the concentration based on the molar definition in settings.

        This method updates the concentration value in the settings dictionary based on the 'Molar definition' key. It calculates concentration values differently based on whether the molar definition is set to 'Molecules', 'Unit cells', or 'Atoms'. It also enables or disables the `natoms_sb` spin box accordingly.
        """        
        debugger.print("Start:: set_concentration")
        if self.settings["Molar definition"] == "Molecules":
            self.settings["concentration"] = 1000.0 / (avogadro_si * self.reader.volume * 1.0e-24 * self.settings["Number of atoms"] / self.reader.nions)
            self.natoms_sb.setEnabled(True)
        elif self.settings["Molar definition"] == "Unit cells":
            self.settings["concentration"]      = 1000.0 / (avogadro_si * self.reader.volume * 1.0e-24)
            self.settings["cell concentration"] = 1000.0 / (avogadro_si * self.reader.volume * 1.0e-24)
            self.natoms_sb.setEnabled(False)
        elif self.settings["Molar definition"] == "Atoms":
            self.settings["concentration"] = 1000.0 / (avogadro_si * self.reader.volume * 1.0e-24 / self.reader.nions)
            self.natoms_sb.setEnabled(False)
        debugger.print("Finished:: set_concentration")
        return

    def writeSpreadsheet(self):
        """Update and write the results of powder and crystal scenarios to a spreadsheet.

        This function navigates through each scenario defined in the `notebook` attribute, extracts relevant data such as absorption coefficients, permittivities, and reflectances, and writes these along with scenario settings to a spreadsheet. It handles different types of scenarios (Powder or Crystal) and makes use of the spreadsheet object's methods for selecting worksheets, writing rows, and dealing with data transformation. The function also handles the calculation of molar absorption coefficients with unit conversion when necessary.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """        
        debugger.print("Start::writeSpreadsheet")
        if self.notebook.spreadsheet is None:
            debugger.print("Finished::writeSpreadsheet spreadsheet is None")
            return
        # make sure the plottingTab is up to date
        self.refresh()
        # Handle powder plots
        molarAbsorptionCoefficients = []
        absorptionCoefficients      = []
        realPermittivities          = []
        imagPermittivities          = []
        sp_atrs                     = []
        R_ps                        = []
        R_ss                        = []
        T_ps                        = []
        T_ss                        = []
        A_ps                        = []
        A_ss                        = []
        powder_legends              = []
        crystal_legends             = []
        # Deal with Scenarios 
        sp = self.notebook.spreadsheet
        sp.selectWorkSheet("Scenarios")
        sp.delete()
        sp.writeNextRow(["A list of the scenarios used:"],col=1)
        for index,scenario in enumerate(self.notebook.scenarios):
            if scenario.scenarioType == "Powder":
                direction = scenario.direction
                depolarisation = scenario.depolarisation
                sp.writeNextRow([""],col=1)
                sp.writeNextRow(["Scenario "+str(index)],col=1,check=1)
                settings = scenario.settings
                for key in sorted(settings,key=str.lower):
                    sp.writeNextRow([key, settings[key]],col=1,check=1)
                sp.writeNextRow(["Normalised unique direction"]+direction.tolist(), col=1,check=1)
                sp.writeNextRow(["Depolarisation matrix"], col=1,check=1)
                sp.writeNextRow(depolarisation[0].tolist(), col=2, check=1)
                sp.writeNextRow(depolarisation[1].tolist(), col=2, check=1)
                sp.writeNextRow(depolarisation[2].tolist(), col=2, check=1)
                molarAbsorptionCoefficients.append( scenario.get_result(self.vs_cm1,self.plot_types[0] ) )
                absorptionCoefficients.append( scenario.get_result(self.vs_cm1,self.plot_types[1] ) )
                realPermittivities.append( scenario.get_result(self.vs_cm1,self.plot_types[2] ) )
                imagPermittivities.append( scenario.get_result(self.vs_cm1,self.plot_types[3] ) )
                sp_atrs.append( scenario.get_result(self.vs_cm1,self.plot_types[4] ) )
                powder_legends.append(scenario.settings["Legend"])
            else:
                sp.writeNextRow([""],col=1)
                sp.writeNextRow(["Scenario "+str(index)],col=1,check=1)
                settings = scenario.settings
                for key in sorted(settings,key=str.lower):
                    sp.writeNextRow([key, settings[key]],col=1,check=1)
                dielectricLayerIndex = scenario.getDielectricLayerIndex()
                if dielectricLayerIndex is not None and scenario.layers[dielectricLayerIndex].isTensor():
                    sp.writeNextRow("Dielectric layer laboratory frame:")
                    sp.writeNextRow(scenario.layers[dielectricLayerIndex].labframe[0].tolist(), col=2, check=1)
                    sp.writeNextRow(scenario.layers[dielectricLayerIndex].labframe[1].tolist(), col=2, check=1)
                    sp.writeNextRow(scenario.layers[dielectricLayerIndex].labframe[2].tolist(), col=2, check=1)
                # Store the reflectance and transmittance
                R_ps.append( scenario.get_result(self.vs_cm1,self.plot_types[5] ) )
                R_ss.append( scenario.get_result(self.vs_cm1,self.plot_types[6] ) )
                T_ps.append( scenario.get_result(self.vs_cm1,self.plot_types[7] ) )
                T_ss.append( scenario.get_result(self.vs_cm1,self.plot_types[8] ) )
                A_ps.append( scenario.get_result(self.vs_cm1,self.plot_types[9] ) )
                A_ss.append( scenario.get_result(self.vs_cm1,self.plot_types[10] ) )
                crystal_legends.append(scenario.settings["Legend"])
        # Single crystal Permittivity
        dielecv = self.notebook.settingsTab.getCrystalPermittivity(self.vs_cm1)
        # Powder results
        # Work out what molar units we are using
        if len(molarAbsorptionCoefficients) > 0:
            if self.settings["Molar definition"] == "Molecules":
                sheet_name = "Powder Molar Absorption (mols)"
            elif self.settings["Molar definition"] == "Unit cells":
                sheet_name = "Powder Molar Absorption (cells)"
            elif self.settings["Molar definition"] == "Atoms":
                sheet_name = "Powder Molar Absorption (atoms)"
            # Always write out the moles of cell
            self.write_powder_results(sp, "Powder Molar Absorption (cells)", self.vs_cm1, powder_legends, molarAbsorptionCoefficients)
            if self.settings["Molar definition"] != "Unit cells":
                # If some other molar definition has been used then write that out too
                molarAbsorptionCoefficients_mols = []
                molar_scaling = self.settings["cell concentration"]/self.settings["concentration"]
                for absorption in molarAbsorptionCoefficients:
                    molarAbsorptionCoefficients_mols.append(molar_scaling * np.array(absorption))
                self.write_powder_results(sp, sheet_name,                      self.vs_cm1, powder_legends, molarAbsorptionCoefficients_mols)
            # end if
            self.write_powder_results(sp, "Powder Absorption",             self.vs_cm1, powder_legends, absorptionCoefficients)
            self.write_powder_results(sp, "Powder Real Permittivity",      self.vs_cm1, powder_legends, realPermittivities)
            self.write_powder_results(sp, "Powder Imaginary Permittivity", self.vs_cm1, powder_legends, imagPermittivities)
            self.write_powder_results(sp, "Powder ATR Reflectance",        self.vs_cm1, powder_legends, sp_atrs)
        # Single Crystal results
        if len(R_ps) > 0:
            self.write_crystal_results(sp, "Crystal R_p", self.vs_cm1, crystal_legends, R_ps)
            self.write_crystal_results(sp, "Crystal R_s", self.vs_cm1, crystal_legends, R_ss)
            self.write_crystal_results(sp, "Crystal T_p", self.vs_cm1, crystal_legends, T_ps)
            self.write_crystal_results(sp, "Crystal T_s", self.vs_cm1, crystal_legends, T_ss)
            self.write_crystal_results(sp, "Crystal A_p", self.vs_cm1, crystal_legends, A_ps)
            self.write_crystal_results(sp, "Crystal A_s", self.vs_cm1, crystal_legends, A_ss)

        if len(dielecv) > 0:
            self.write_eps_results(sp, self.vs_cm1, dielecv)
        debugger.print("Finished::writeSpreadsheet")
        return

    def write_eps_results(self, sp, vs, dielecv):
        """Write real and imaginary parts of crystal permittivity to a spreadsheet.

        Parameters
        ----------
        sp : object
            An instance of a spreadsheet processing class with methods to manipulate data in a spreadsheet.
        vs : list
            A list containing frequency values.
        dielecv : numpy.ndarray
            A complex numpy array where the real parts represent the real permittivities and the imaginary parts represent the imaginary permittivities of a crystal. The array should have a shape of (N,3,3) where N is the number of frequency values, and the 3x3 inner arrays represent the permittivity tensor for each frequency.

        Returns
        -------
        None

        Notes
        -----
        The function does two main tasks:
        1. Selects the 'Real Crystal Permittivity' sheet, deletes its current content if any, and writes the real parts of the permittivity tensor for each frequency along with the frequency values themselves.
        2. Selects the 'Imag Crystal Permittivity' sheet, deletes its current content if any, and writes the imaginary parts of the permittivity tensor for each frequency along with the frequency values themselves.

        Both sections write data in the format: frequencies (cm-1), xx, yy, zz, xy, xz, yz, where xx, yy, zz, xy, xz, and yz are components of the permittivity tensor.
        The output data starts from the second column (index 1), and for each row written, a 'check' flag is set to 1.

        """        
        debugger.print("Start:: write_eps_results length vs",len(vs))
        sp.selectWorkSheet("Real Crystal Permittivity")
        sp.delete()
        headers = ["frequencies (cm-1)", "xx", "yy", "zz", "xy", "xz", "yz" ]
        sp.writeNextRow(headers,row=0, col=1)
        for v,eps in zip(vs,dielecv):
            eps_xx_r = np.real(eps[0][0])
            eps_yy_r = np.real(eps[1][1])
            eps_zz_r = np.real(eps[2][2])
            eps_xy_r = np.real(eps[0][1])
            eps_xz_r = np.real(eps[0][2])
            eps_yz_r = np.real(eps[1][2])
            output = [v, eps_xx_r, eps_yy_r, eps_zz_r, eps_xy_r, eps_xz_r, eps_yz_r ]
            sp.writeNextRow(output, col=1,check=1)
        sp.selectWorkSheet("Imag Crystal Permittivity")
        sp.delete()
        sp.writeNextRow(headers,row=0, col=1)
        for v,eps in zip(vs,dielecv):
            eps_xx_i = np.imag(eps[0][0])
            eps_yy_i = np.imag(eps[1][1])
            eps_zz_i = np.imag(eps[2][2])
            eps_xy_i = np.imag(eps[0][1])
            eps_xz_i = np.imag(eps[0][2])
            eps_yz_i = np.imag(eps[1][2])
            output = [v, eps_xx_i, eps_yy_i, eps_zz_i, eps_xy_i, eps_xz_i, eps_yz_i ]
            sp.writeNextRow(output, col=1,check=1)
        debugger.print("Finished:: write_eps_results length vs",len(vs))
        return

    def write_crystal_results(self, sp, name, vs, legends, yss):
        """Write single crystal results to a spread sheet.

        Parameters
        ----------
        sp : object
            The spreadsheet object.
        name : str
            The worksheet name used for writing.
        vs : np.array
            An array of the frequencies.
        legends : list
            The heading names for the yss.
        yss : list of np.arrays
            A list of numpy arrays of the reflections and transmittance.

        Returns
        -------
        None

        """
        debugger.print("Start:: write_crystal_results")
        debugger.print("write_crystal_results name",name)
        debugger.print("write_crystal_results legends",legends)
        debugger.print("write_crystal_results length vs",len(vs))
        sp.selectWorkSheet(name)
        sp.delete()
        headers = ["frequencies (cm-1)"]
        headers.extend(legends)
        sp.writeNextRow(headers,row=0, col=1)
        for iv,v in enumerate(vs):
           output = [v]
           for ys in yss:
               output.append(ys[iv])
           sp.writeNextRow(output, col=1,check=1)
        debugger.print("Finished:: write_crystal_results")
        return

    def write_powder_results(self, sp, name, vs, legends, yss):
        """Write the powder simulation results to a worksheet.

        Parameters
        ----------
        sp : object
            An object with methods for manipulating a worksheet.
        name : str
            The name of the worksheet to select or create.
        vs : list
            A list of frequency values (in cm-1).
        legends : list of str
            The legends to be used as headers, alongside 'frequencies (cm-1)'.
        yss : list of lists
            A list where each element is a list of intensities corresponding to each frequency in `vs`.

        Returns
        -------
        None

        """        
        debugger.print("Start:: write powder results")
        debugger.print("write_powder_results name",name)
        debugger.print("write_powder_results legends",legends)
        debugger.print("write_powder_results length vs",len(vs))
        sp.selectWorkSheet(name)
        sp.delete()
        headers = ["frequencies (cm-1)"]
        #for isc,ys in enumerate(yss):
        #    headers.append('Scenario'+str(isc))
        headers.extend(legends)
        sp.writeNextRow(headers,row=0, col=1)
        for iv,v in enumerate(vs):
           output = [v]
           for ys in yss:
               output.append(ys[iv])
           sp.writeNextRow(output, col=1,check=1)
        debugger.print("Finished:: write powder results")
        return

    def plot(self):
        # import matplotlib.pyplot as pl
        # mp.use('qtagg')
        """Plot the results based on specified settings and scenarios.

        This function generates a plot for the given scenarios under the analysis settings
        defined in the notebook object's tabs. It handles various conditions to properly format
        the plot, including checking if necessary parameters and objects are set, converting
        frequency units, and adjusting data for display based on plot type. The actual plotting
        is done using matplotlib, and various attributes like plot legends, titles, and axes labels
        are dynamically set based on the configuration.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        - self.notebook.mainTab.settings: Dictionary containing program settings.
        - self.notebook.mainTab.getFullFileName(): Method that returns the currently selected filename.
        - self.notebook.mainTab.reader: Object used for reading data files.
        - self.notebook.settingsTab.CrystalPermittivityObject: Object containing crystal permittivity settings.
        - self.settings: Dictionary containing plot-related settings such as frequency range and plot type.
        - self.notebook.scenarios: List of scenarios to be plotted.

        """        
        debugger.print("Start:: plot")
        # Assemble the mainTab settings
        settings = self.notebook.mainTab.settings
        program = settings["Program"]
        filename = self.notebook.mainTab.getFullFileName()
        reader = self.notebook.mainTab.reader
        if reader is None:
            debugger.print("Finished:: plot aborting because reader is NONE")
            return
        if program == "":
            debugger.print("Finished:: plot aborting because program is not set")
            return
        if filename == "":
            debugger.print("Finished:: plot aborting because filename is not set")
            return
        if self.notebook.settingsTab.CrystalPermittivityObject is None:
            debugger.print("Finished:: plot aborting because settingTab.CrystalPermittivityObject is not set")
            return
        QApplication.setOverrideCursor(Qt.WaitCursor)
        vmin = self.settings["Minimum frequency"]
        vmax = self.settings["Maximum frequency"]
        vinc = self.settings["Frequency increment"]
        self.vs_cm1 = np.arange(float(vmin), float(vmax)+0.5*float(vinc), float(vinc))
        self.subplot = None
        self.figure.clf()
        x = np.array(self.vs_cm1)
        removeFirstElement = False
        reverseElements = False
        if x[0] <= 0:
            # We can't have a 0 frequency converted to a wavelength
            removeFirstElement = True
            x = x[1:]
        if isThisAFrequencyUnit(self.settings["Frequency unit"]):
            x = x[::-1]
            reverseElements = True
        xlabel = self.settings["Frequency unit"]
        if self.settings["Frequency unit"] == "wavenumber":
            xlabel = r"Frequency $\mathdefault{(cm^{-1})}$"
        x = convert_frequency_units( x, "wavenumber", self.settings["Frequency unit"] )
        self.subplot = self.figure.add_subplot(111)
        self.notebook.progressbars_set_maximum(self.get_total_number_of_frequency_calculations())
        self.legends = []
        plots = 0
        for scenario in self.notebook.scenarios:
            legend = scenario.settings["Legend"]
            self.legends.append(legend)
            y = scenario.get_result(self.vs_cm1,self.settings["Plot type"])
            if y is not None and len(y) > 0:
                y = np.array(y)
                if removeFirstElement:
                    y = y[1:]
                if reverseElements:
                     y = y[::-1]
                if self.settings["Plot type"] == "Powder Molar Absorption":
                    y = y * self.settings["cell concentration"]/self.settings["concentration"]
                plots += 1
                line, = self.subplot.plot(x,y,lw=2, label=legend )
        if plots > 0:
            self.subplot.set_xlabel(xlabel)
            self.subplot.set_ylabel(self.plot_ylabels[self.settings["Plot type"]])
            self.subplot.legend(loc="best")
            self.subplot.set_title(self.settings["Plot type"])
            #self.subplot.set_autoscaley_on(False)
            #self.subplot.set_ylim([0,1])
            self.canvas.draw_idle()
        QApplication.restoreOverrideCursor()
        debugger.print("Finished:: plot")

    def get_number_of_calculations_required(self):
        """Return the total number of spectra that need to be calculated.

        Only spectra that need a refresh are included.

        Parameters
        ----------
        None

        Returns
        -------
        int
            The total number of spectra that require calculation.

        """
        debugger.print("Start:: get_number_of_calculations_required")
        n = 0
        for scenario in self.notebook.scenarios:
            n += scenario.getNoCalculationsRequired()
        debugger.print("get_number_of_calculations_required",n)
        return n

    def greyed_out(self):
        """Handle items that should be greyed out if they are not needed.

        Parameters
        ----------
        None

        Returns
        -------
        int

        """
        debugger.print("Start:: greyed_out")
        powder_scenarios_present = False
        crystal_scenarios_present = False
        for scenario in self.notebook.scenarios:
            if scenario.scenarioType == "Powder":
                powder_scenarios_present = True
            else:
                crystal_scenarios_present = True
        # end of for loop
        # 
        # Disable any plot types that are not needed
        #
        self.plot_type_cb.model().item(0).setEnabled(True)
        self.plot_type_cb.model().item(1).setEnabled(True)
        self.plot_type_cb.model().item(2).setEnabled(True)
        self.plot_type_cb.model().item(3).setEnabled(True)
        self.plot_type_cb.model().item(4).setEnabled(True)
        self.plot_type_cb.model().item(5).setEnabled(True)
        self.plot_type_cb.model().item(6).setEnabled(True)
        self.plot_type_cb.model().item(7).setEnabled(True)
        self.plot_type_cb.model().item(8).setEnabled(True)
        self.plot_type_cb.model().item(9).setEnabled(True)
        self.plot_type_cb.model().item(10).setEnabled(True)
        index = self.plot_type_cb.findText(self.settings["Plot type"], Qt.MatchFixedString)
        if not powder_scenarios_present:
            self.plot_type_cb.model().item(0).setEnabled(False)
            self.plot_type_cb.model().item(1).setEnabled(False)
            self.plot_type_cb.model().item(2).setEnabled(False)
            self.plot_type_cb.model().item(3).setEnabled(False)
            self.plot_type_cb.model().item(4).setEnabled(False)
            if index < 5:
                self.plot_type_cb.setCurrentIndex(5)
                self.settings["Plot type"] = self.plot_type_cb.currentText()
        if not crystal_scenarios_present:
            self.plot_type_cb.model().item(5).setEnabled(False)
            self.plot_type_cb.model().item(6).setEnabled(False)
            self.plot_type_cb.model().item(7).setEnabled(False)
            self.plot_type_cb.model().item(8).setEnabled(False)
            self.plot_type_cb.model().item(9).setEnabled(False)
            self.plot_type_cb.model().item(10).setEnabled(False)
            if index >= 5:
                self.plot_type_cb.setCurrentIndex(0)
                self.settings["Plot type"] = self.plot_type_cb.currentText()
        debugger.print("Finished:: greyed_out")



#!/usr/bin/env python
#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""The checkexcel package reads in two excel files and check that the float contents are the same.

It is used by the :mod:`~PDielec.pdmake` command to check the validity of the reference calculations in the test suite.

"""
import sys

from openpyxl import load_workbook
from termcolor import colored

global threshold

def main():

    # Start processing the directories
    """Compare two Excel files for any significant numerical changes and report discrepancies.

    This script takes in two Excel files as arguments and optionally,
    a threshold for numerical comparison and a flag for full sheet comparison.

    Full comparison includes 'Settings' and 'Scenarios' sheets by default excluded.

    Parameters
    ----------
    None directly to the function. Parses command line arguments:
    - Two file paths.
    - Optional `-thresh` followed by a float value for the threshold of numerical difference (default is 1.0E-3).
    - Optional `-f` flag for forcing a full comparison including typically excluded sheets.

    Returns
    -------
    tuple
        A tuple containing the following elements:
        - Number of errors (`nerrors`): int
        - Row of the last error or discrepancy (`row`): int
        - Column of the last error or discrepancy (`col`): int
        - Sheet name where the last error or discrepancy was found (`sheet`): str
        - First file name (`file1`): str
        - Value from the first file associated with the last discrepancy (`value1`): float or str
        - Second file name (`file2`): str
        - Value from the second file associated with the last discrepancy (`value2`): float or str
        - Maximum percentage error found during the comparison (`max_percentage_error`): float

    Raises
    ------
    Depends on the Excel library used, typically:
    - `openpyxl.utils.exceptions.InvalidFileException` if an invalid file or path is provided.
    - Other exceptions related to file I/O or processing might be raised depending on the content and structure of the input Excel files.

    Notes
    -----
    - The script will immediately exit with usage instructions if less than two file paths are provided.
    - Numerical difference is calculated only for numeric data. For text data, a simplified equivalence check is done.
    - The script output includes printing to standard error for usage, errors, or status, with optional ANSI color highlighting.

    """    
    if len(sys.argv) <= 1 :
        print("checkexcel file file2 [-thresh 1.0E-3] [-f]", file=sys.stderr)
        print("         Compare the two excel files for any significant changes", file=sys.stderr)
        print("         Numbers f1 and f2 in each file are compared", file=sys.stderr)
        print("         an error is flagged if 2*abs(f1-f2)/(f1+f2+2) > threshold", file=sys.stderr)
        print("         The default value for threshold is 1.0E-3 ", file=sys.stderr)
        print("         -f forces a full comparison of the sheets ", file=sys.stderr)
        print("            by default Settings and Scenarios are not included    ", file=sys.stderr)
        sys.exit()

    threshold = 1.0E-3
    tokens = sys.argv[1:]
    ntokens = len(tokens)-1
    itoken = -1
    files = []
    full = False
    while itoken < ntokens:
        itoken += 1
        token = tokens[itoken]
        if token == "-thresh":
            itoken +=1
            threshold = float(tokens[itoken])
        if token == "-f":
            full = True
        else:
            files.append(tokens[itoken])
        # end if
    # end while

    file1 = files[0]
    file2 = files[1]

    #
    # Open the work books
    #
    wb1 = load_workbook(filename=file1, read_only=True)
    wb2 = load_workbook(filename=file2, read_only=True)
    #
    # Initialise variables
    #
    error = None
    nerrors = 0
    max_percentage_error = 0.0
    row = 0
    col = 0
    sheet = ""
    file1 = "" 
    value1 = 0
    file2 = "" 
    value2 = 0
    #
    # Loop over sheets
    #
    sheets = ["Powder Molar Absorption (cells)","Powder Absorption","Powder Real Permittivity","Powder Imaginary Permittivity", "Powder ATR Reflectance", "Analysis","Crystal R_p","Crystal R_s","Crystal T_p","Crystal T_s","Real Crystal Permittivity","Imag Crystal Permittivity"]
    if full:
        sheets.append("Main")
        sheets.append("Settings")
        sheets.append("Scenarios")
    for sheet in sheets:
        if sheet not in wb1 :
            continue
        if sheet not in wb2 :
            continue
        print("Checking sheet ",sheet)
        ws1 = wb1[sheet]
        ws2 = wb2[sheet]
        max_rows1 = ws1.max_row
        max_rows2 = ws2.max_row
        max_columns1 = ws1.max_column
        max_columns2 = ws2.max_column
        if max_rows1 != max_rows2:
            print("Error - the number rows in the sheet are not the same",sheet,max_rows1,max_rows2)
            nerrors += 1
            continue
        if max_columns1 != max_columns2:
            print("Error - the number columns in the sheet are not the same",sheet,max_columns1,max_columns2)
            nerrors += 1
            continue
        #
        # Loop over rows
        #
        for row_index,(row1,row2) in enumerate(zip(ws1.rows, ws2.rows)):
            #
            # Loop over cells
            #
            for col_index,(cell1,cell2) in enumerate(zip(row1,row2)):
                value1 = cell1.value
                value2 = cell2.value
                if cell1 is None and cell2 is None:
                    pass
                elif cell1 is None or cell2 is None:
                    nerrors += 1
                    percentage_error = 0.0
                    error = (sheet, row_index, col_index, value1, value2, percentage_error)
                elif value1 != value2:
                    if value1 is not None and value2 is not None and cell1.data_type == "n" and cell2.data_type == "n":
                        #
                        # Flag an error which is numeric
                        #
                        percentage_error = 100.0*abs(2.0*(value1 - value2)/(abs(value1) + abs(value2)+2))
                        if percentage_error > 100.*threshold:
                            nerrors += 1
                            if percentage_error > max_percentage_error:
                                max_percentage_error = percentage_error
                                error = (sheet, row_index, col_index, value1, value2, percentage_error)
                            # if percentage error
                        elif percentage_error > max_percentage_error:
                            max_percentage_error = percentage_error
                        # if percentage error > threshold
                    else:
                        #
                        # This is a non-numeric error
                        # Remove any back or forward slashes to avoid problems with filenames in linux/windows
                        #
                        if isinstance(value1,str):
                            value1 = value1.replace("\\","")
                            value1 = value1.replace("/","")
                        if isinstance(value2,str):
                            value2 = value2.replace("\\","")
                            value2 = value2.replace("/","")
                        if value1 != value2:
                            nerrors += 1
                            error = (sheet, row_index, col_index, value1, value2, 0.0)
                        # if cell1.data_type
                    # if value1 != value2
                # if cell1 is none
            # for cell1, cell2
        # for rowq, row
    # for sheet
    if error is not None:
        sheet,row,col,value1,value2,max_percentage_error = error
        print("  "+colored("ERRORS:","red")+f"({nerrors}) LARGEST ON ROW,COL {row},{col} OF SHEET {sheet}, {file1}({value1}) and {file2}({value2}) -- max %error={max_percentage_error}")
    elif nerrors > 0:
        print("  "+colored("ERRORS:","red")+f"({nerrors}) Dimensions of spreadsheet were wrong                                    ")
    else:
        print("  "+colored("OK:","blue")+f" {file1} = {file2} -- max %error={max_percentage_error}")
    return nerrors, row,col,sheet,file1,value1,file2,value2,max_percentage_error

if __name__=="__main__":
    main()



#
# Copyright 2024 John Kendrick & Andrew Burnett
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License along with this program, if not see https://opensource.org/licenses/MIT
#
"""SingleCrystalLayer module."""
# -*- coding: utf8 -*-
import sys

import numpy as np
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QSizePolicy,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from PDielec.Utilities import Debug


class SingleCrystalLayer:
    """A class representing a single crystal layer.

    Parameters
    ----------
    material : Material
        An instance of Material object (see :class:`~PDielec.Materials.Material`), which contains:
          - name (str)
          - density (float)
          - permittivity (float)
          - unit cell (UnitCell)
    hkl : list of int
        A list of 3 integers defining the plane of interest.
    azimuthal : float
        The azimuthal angle of rotation of the crystal about the z-axis.
    thickness : float
        The thickness of the layer in the specified thickness units.
    thicknessUnits : {'ang','nm', 'um', 'mm', 'cm'}
        The units of thickness measurement.
    incoherentOption : {'Coherent', 'Incoherent (intensity)', 'Incoherent (phase cancelling)', 'Incoherent (phase averaging)', 'Incoherent non-reflective'}
        The option for handling incoherent scattering. 
    dielectricFlag : bool
        True if the layer material is the dielectric being studied (ie it was read in as a DFT calculation), False otherwise.

    Attributes
    ----------
    material : Material
        The material of the layer. (See :class:`~PDielec.Materials.Material`)
    hkl : list of int
        The Miller indices of the plane of interest in the crystal.
    incoherentOption : str
        The incoherent option for the layer.
    azimuthal : float
        The azimuthal angle of rotation of the crystal about the z-axis, in degrees.
    thickness : float
        The thickness of the layer in the specified units.
    thicknessUnit : str
        The units used for the thickness measurement.
    dielectricFlag : bool
        A flag indicating if the layer material is the dielectric material.
    euler : ndarray
        The Euler rotation matrix for the crystal to laboratory frame transformation, stored as a 3x3 numpy array with `dtype=np.longdouble`.
    euler_inverse : ndarray
        The inverse of the Euler rotation matrix, stored as a 3x3 numpy array with `dtype=np.longdouble`.
    labframe_w : 
        Placeholder attribute for GUI interactions, not initialized within the class definition.
    labframe : ndarray
        The lattice vectors of the crystal in the laboratory frame, stored as a numpy array.
    phaseShift : float
        The phase shift for the layer.

    Methods
    -------
    print()
        Prints the main attributes of the layer.
    getPhaseShift()
        Returns the phase shift for the layer.
    setPhaseShift(phaseShift)
        Sets the phase shift for the layer.
    setAzimuthal(angle)
        Sets the azimuthal angle of the layer and recalculates the Euler matrix.
    isCoherent()
        Returns True if the layer is coherent, False otherwise.
    getIncoherentOption()
        Returns the incoherent interaction option for the layer.
    setIncoherentOption(option)
        Sets the incoherent interaction option for the layer.
    getAzimuthal()
        Returns the azimuthal angle.
    setThickness(thickness)
        Sets the thickness of the layer.
    getThickness()
        Returns the thickness of the layer.
    getThicknessInMetres()
        Converts and returns the thickness of the layer in meters.
    setThicknessUnit(thicknessUnit)
        Sets the unit of thickness measurement.
    getThicknessUnit()
        Returns the unit of thickness measurement.
    setHKL(hkl)
        Sets the Miller indices for the plane of interest in the crystal.
    getHKL()
        Returns the Miller indices of the plane of interest.
    getName()
        Returns the name of the material.
    getMaterial()
        Returns the material object associated with the layer.
    getPermittivityFunction()
        Returns the permittivity tensor for the crystal at the surface and azimuthal angle specified.
    isTensor()
        Returns True if the material of the layer is a tensor material, False otherwise.
    isScalar()
        Returns True if the material of the layer is a scalar material, False otherwise.
    isDielectric()
        Returns True if the material of the layer is the dielectric being studied.
    calculate_euler_matrix()
        Calculates the Euler angles for the crystal to lab frame transformation.
    azimuthalRotationMatrix(angle)
        Calculates a rotation matrix for the azimuthal angle.
    changeLabFrameInfo()
        Changes the Lab Frame Information, used with GUI interactions.
    getLabFrame()
        Returns the information of the crystal axes in the laboratory frame.
    calculateLabFrameEpsilon(v)
        Computes the permittivity tensor for the material at a given frequency and transforms it to the laboratory frame.
    invert(m)
        Calculates the inverse of a given 3x3 complex matrix.

    Notes
    -----
    The class heavily relies on numpy for various calculations including matrix manipulations and rotation calculations.

    """

    def __init__(self,material,hkl=None,azimuthal=0.0,thickness=0.0,thicknessUnit="nm",incoherentOption="Coherent",dielectricFlag=False):
        """Initialise a single crystal layer.

        Parameters
        ----------
        material : Material
            An instance of :class:`~PDielec.Materials.Material`, which contains:
              - name (str)
              - density (float)
              - permittivity (float)
              - unit cell (UnitCell)
        hkl : list of ints
            A list of 3 integers defining the plane of interest. Default is [0,0,1]
            If the material is isotropic then [0,0,0] is used
        azimuthal : float
            The azimuthal angle of rotation of the crystal about the z-axis.
        thickness : float
            The thickness of the layer in the specified thickness units.
        thicknessUnit : {'ang','nm', 'um', 'mm', 'cm'}
            The units of thickness measurement.
        incoherentOption : {'Coherent', 'Incoherent (intensity)', 'Incoherent (phase cancelling)', 'Incoherent (phase averaging)', 'Incoherent non-reflective'}
            The option for handling incoherent scattering. 
        dielectricFlag : bool
            True if the layer material is the dielectric being studied (ie it was read in as a DFT calculation), False otherwise.

        """
        if hkl is None:
            hkl = [0,0,1]
        self.material = material
        if material.isScalar():
            hkl = [0,0,0]
        self.hkl = hkl
        self.incoherentOption = incoherentOption
        self.azimuthal = azimuthal
        self.thickness = thickness
        self.thicknessUnit = thicknessUnit
        self.dielectricFlag = dielectricFlag
        self.euler = np.zeros((3,3),dtype=np.longdouble)
        self.euler_inverse = np.zeros((3,3),dtype=np.longdouble)
        self.labframe_w = None
        self.labframe = None
        self.phaseShift = 0.0
        self.calculate_euler_matrix()

    def print(self):
        """Print the main attributes of the layer.

        Parameters
        ----------
        None

        """
        print("------------------ ")
        print("Material         : ", self.material)
        print("HKL              :      ", self.hkl)
        print("Azimuthal        :", self.azimuthal)
        print("Thickness        :", self.thickness)
        print("Thickness unit   :", self.thicknessUnit)
        print("Dielectric flag  :", self.dielectricFlag)
        print("Incoherent option:", self.incoherentOption)
        print("Phase shift      :", self.phaseShift)

    def getPhaseShift(self):
        """Get the phase shift for the layer.

        Parameters
        ----------
        None

        Returns
        -------
        float
            The phase shift used in the averaged phase shift method

        """
        return self.phaseShift

    def setPhaseShift(self, phaseShift):
        """Set the phase shift for the layer.

        Parameters
        ----------
        phaseShift : float
           The phase shift

        Returns
        -------
        None

        """
        self.phaseShift = phaseShift
        return

    def setAzimuthal(self, angle):
        """Set the azimuthal angle of the layer and calculate the Euler matrix.

        Parameters
        ----------
        angle : float
           The azimuthal angle in degrees

        Returns
        -------
        None

        """
        self.azimuthal = angle
        self.calculate_euler_matrix()
        return 

    def isCoherent(self):
        """Return True if this is a coherent layer, False otherwise.

        Parameters
        ----------
        None

        Returns
        -------
        bool
            True if this is a coherent layer, False otherwise.

        """
        result = False
        if self.incoherentOption == "Coherent":
            result = True
        return result

    def getIncoherentOption(self):
        """Return the value of the incoherent option attribute.

        This will return 'Coherent' for a coherent layer

        Parameters
        ----------
        None

        Returns
        -------
        type
            The value of the `incoherentOption` attribute of the object.

        """        
        return self.incoherentOption

    def setIncoherentOption(self,option):
        """Set the incoherent option.

        Parameters
        ----------
        option : string
            Sets the incoherent option.  Can be one of;
            - "Coherent"
            - "Incoherent (intensity)"
            - "Incoherent (phase averaging)"
            - "Incoherent (phase cancelling)"
            - "Incoherent (non-reflective)"

        Returns
        -------
        None

        """
        self.incoherentOption = option
        return 

    def getAzimuthal(self):
        """Get the azimuthal angle.

        Parameters
        ----------
        None

        Returns
        -------
        float
            The azimuthal angle.

        """
        return self.azimuthal

    def setThickness(self, thickness):
        """Set the thickness.

        Parameters
        ----------
        thickness : float
            The value of the layer thickness in the current thickness units

        Returns
        -------
        None

        """
        self.thickness = thickness
        return 

    def getThickness(self):
        """Get the thickness.

        Parameters
        ----------
        None

        Returns
        -------
        float

        """
        return self.thickness

    def getThicknessInMetres(self):
        """Get the thickness in metres.

        Parameters
        ----------
        None

        Returns
        -------
        float

        """
        thickness_conversion_factors = {"ang":1.0E-10, "nm":1.0E-9, "um":1.0E-6, "mm":1.0E-3, "cm":1.0E-2}
        tom = thickness_conversion_factors[self.thicknessUnit]
        return tom*self.thickness

    def setThicknessUnit(self, thicknessUnit):
        """Set the thickness unit.

        Parameters
        ----------
        thicknessUnit : str
            The unit of thickness. Can be one of "ang", "nm", "um", "mm", or "cm"

        Returns
        -------
        None

        """
        self.thicknessUnit = thicknessUnit
        return 

    def getThicknessUnit(self):
        """Get the thickness unit.

        Parameters
        ----------
        None

        Returns
        -------
        str

        """
        return self.thicknessUnit

    def setHKL(self, hkl):
        """Set the hkl and recalculate the Euler matrix.

        Parameters
        ----------
        hkl : list of 3 ints
           The hkl parameters of the surface

        Returns
        -------
        None

        """
        self.hkl = hkl
        self.calculate_euler_matrix()
        return 

    def getHKL(self):
        """Get the hkl.

        Parameters
        ----------
        None

        Returns
        -------
        list of 3 ints

        """
        return self.hkl

    def getName(self):
        """Return the material name.

        Parameters
        ----------
        None

        Returns
        -------
        str
            The name of the material.

        """
        return self.material.getName()

    def getMaterial(self):
        """Return the material.

        Parameters
        ----------
        None

        Returns
        -------
        an instance of a Materials object

        """
        return self.material

    def getPermittivityFunction(self):
        """Return the permittivity function associated with this material.

        Parameters
        ----------
        None

        Returns
        -------
        A permittivity function

        """
        return self.calculateLabFrameEpsilon

    def isTensor(self):
        """Return true if the material of the layer is a tensor material.

        Parameters
        ----------
        None

        Returns
        -------
        bool
            True if the material of the layer is a tensor material, False otherwise.

        """
        return self.getMaterial().isTensor()

    def isScalar(self):
        """Return true if the material of the layer is a scalar material.

        Parameters
        ----------
        None

        Returns
        -------
        bool
            True if the material of the layer is a scalar material, False otherwise.

        """
        return self.getMaterial().isScalar()

    def isDielectric(self):
        """Return true if the material of the layer is the dielectric material.

        Parameters
        ----------
        None

        Returns
        -------
        bool
            True if the material of the layer is the dielectric material, otherwise False.

        """
        return self.dielectricFlag

    def calculate_euler_matrix(self):
        """Calculate the Euler angles for the crystal to lab transformation.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        # Get plane specification
        hkl = self.hkl
        sum2 = hkl[0]*hkl[0] + hkl[1]*hkl[1] + hkl[2]*hkl[2]
        if sum2 < 1:
            return 
        x = 0
        y = 1
        z = 2
        # convert normal to plane to a direction in xyz coordinates
        planez = self.material.cell.convert_hkl_to_xyz(hkl)
        planez /=  np.linalg.norm(planez)
        plane = np.zeros( (3,3) )
        lab   = np.identity(3)
        plane[z] = planez
        if plane[z][2] < 0.99999999 and plane[z][2] > -0.99999999:
            plane[x] = np.cross(plane[z], lab[z])
            plane[y] = np.cross(plane[z], plane[x])
            plane[x] /= np.linalg.norm(plane[y])
            plane[y] /= np.linalg.norm(plane[y])
        else:
            plane[x] = lab[x]
            plane[y] = lab[y]
            plane[z] = lab[z]
        # Calculate the rotation matrix which transforms us to a unit matrix
        rotation = np.linalg.pinv(plane)
        # Because the formula used for the Euler transform is based on an active transform
        # We calculate the angles using the transpose of the rotation matrix
        self.euler = rotation.T
        # Rotate by azimuthal angle
        self.euler = np.matmul(self.azimuthalRotationMatrix(self.azimuthal),self.euler)
        # Calculate and keep euler inverse
        self.euler_inverse = self.invert(self.euler)
        # Calculate lab frame
        self.labframe = np.matmul(self.euler,self.material.cell.lattice.T).T
        normal_to_plane_lab = np.matmul(self.euler,plane[z])
        if normal_to_plane_lab[2] < 0.9999 and normal_to_plane_lab[2] > -0.9999:
            print("Error in Euler rotations - surface normal is not along Z-axis", normal_to_plane_lab)
            sys.exit()
        return 

    def azimuthalRotationMatrix(self,angle):
        """Calculate a rotation matrix for the azimuthal angle (in degrees).

        Parameters
        ----------
        angle : float
            The azimuthal angle in degrees

        Returns
        -------
        ndarray 3x3
            The rotation matrix for the given azimuthal angle.

        """
        angle = np.radians(angle)
        matrix = np.eye(3)
        # in numpy the first index is the column, the second is the row
        matrix[0,0] = np.cos(angle)
        matrix[1,1] = np.cos(angle)
        matrix[0,1] = -np.sin(angle)
        matrix[1,0] = +np.sin(angle)
        return matrix

    def changeLabFrameInfo(self):
        """Change the Lab Frame Info panel.

        Information about the relationship between the crystal and labframe coordinate systems is presented.
        """
        if self.labframe_w is None:
            return
        a,b,c = self.getLabFrame()
        self.labframe_w.clear()
        self.labframe_w.addItem(f"crystal a-axis in lab frame: {a[0]: 3.5f}, {a[1]: 3.5f}, {a[2]: 3.5f}" )
        self.labframe_w.addItem(f"crystal b-axis in lab frame: {b[0]: 3.5f}, {b[1]: 3.5f}, {b[2]: 3.5f}" )
        self.labframe_w.addItem(f"crystal c-axis in lab frame: {c[0]: 3.5f}, {c[1]: 3.5f}, {c[2]: 3.5f}" )
        return

    def getLabFrame(self):
        """Return the a, b, and c cell axes in the laboratory frame coordinates.

        Parameters
        ----------
        None

        Returns
        -------
        (a, b, c) : 3 lists of floats each with 3 elements
            - a, The a cell axis in the laboratory frame.
            - b, The b cell axis in the laboratory frame.
            - c, The c cell axis in the laboratory frame.

        """
        return self.labframe

    def calculateLabFrameEpsilon(self, v):
        """Get the permittivity for this material at frequency v and transform it to the laboratory frame.

        Parameters
        ----------
        v : float
            Frequency at which the permittivity is evaluated.

        Returns
        -------
        float
            The permittivity of the material in the laboratory frame.

        """
        epsilon_xstal = self.material.getPermittivityFunction()(v)
        if self.material.isScalar():
            self.labFrameEpsilon = epsilon_xstal * np.eye(3)
        else:
            self.labFrameEpsilon = np.matmul(self.euler, np.matmul(epsilon_xstal,self.euler_inverse))
        return self.labFrameEpsilon

    def invert(self,m):
        """Calculate the inverse of m.

        Parameters
        ----------
        m : complex 3x3 array
            The input matrix for which to calculate the inverse.

        Returns
        -------
        numpy.ndarray
            The inverse of `m` as a 3x3 complex numpy array.

        """
        m1, m2, m3, m4, m5, m6, m7, m8, m9 = m.flatten()
        determinant = m1*m5*m9 + m4*m8*m3 + m7*m2*m6 - m1*m6*m8 - m3*m5*m7 - m2*m4*m9  
        return np.array([[m5*m9-m6*m8, m3*m8-m2*m9, m2*m6-m3*m5],
                         [m6*m7-m4*m9, m1*m9-m3*m7, m3*m4-m1*m6],
                         [m4*m8-m5*m7, m2*m7-m1*m8, m1*m5-m2*m4]])/determinant

class ShowLayerWindow(QDialog):
    """A GUI window for displaying and editing the properties of a layer in a project.

    Parameters
    ----------
    layer : object
        The layer object to be displayed or acted upon in the window. (see :class:`SingleCrystalLayer`)
    message : str, optional
        Custom message to display in the window. The default is an empty string.
    parent : QWidget, optional
        The parent widget of this window. The default is None.
    debug : bool, optional
        Flag to enable or disable debug mode. The default is False.

    Attributes
    ----------
    layer : object
        The :class:`SingleCrystalLayer` instance that is displayed or being edited in this window.
    message : str
        The message displayed at the top of the window.
    layout : QVBoxLayout
        The main layout of the window.
    buttonBox : QDialogButtonBox
        The dialog box containing Ok and Cancel buttons.

    Methods
    -------
    getLayer()
        Return the :class:`SingleCrystalLayer` instance being edited or displayed.
    drawLayerWidget()
        Create a QWidget displaying all information about the layer.
    drawLayerWidgetLine1()
        Add widgets to display the first line of layer properties.
    drawLayerWidgetLine2()
        Add widgets to display the second line of layer properties related to the crystallographic direction (hkl).
    drawLayerWidgetLine3()
        Add widgets for displaying the third line of layer properties related to lab frame information.
    changeLabFrameInfo()
        Update the displayed lab frame information based on the layer's current state.
    on_film_thickness_sb_changed(value)
        Respond to changes in the film thickness spin box.
    on_azimuthal_angle_sb_changed(value)
        Respond to changes in the azimuthal angle spin box.
    on_hkl_sb_changed(value, hkorl)
        Respond to changes in the h, k, or l spin boxes.
    on_thickness_units_cb_activated(index)
        Respond to thickness unit selection changes.

    """

    def __init__(self, layer, message = "", parent=None, debug=False ):
        """Initialize a ShowLayerWindow instance.

        Parameters
        ----------
        layer : object
            The layer object to be displayed or acted upon in the window.
        message : str, optional
            Custom message to display in the window. The default is an empty string.
        parent : QWidget, optional
            The parent widget of this window. The default is None.
        debug : bool, optional
            Flag to enable or disable debug mode. The default is False.

        Notes
        -----
        This constructor initializes the ShowLayerWindow by setting up the UI elements
        including buttons and the layout. It also configures the debug mode according
        to the provided argument.

        """        
        super().__init__(parent)
        global debugger
        debugger = Debug(debug,"ShowLayerWindow")
        debugger.print("Start:: initialiser")
        # Set up the buttons of the button box
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        # Initialize the layer
        self.layer = layer
        self.message = message
        # The dialog will have a vertical layout
        self.layout = QVBoxLayout(self)
        # Create the layer widget to display
        layerWidget = self.drawLayerWidget()
        # Add the widget showing the layer information
        self.layout.addWidget(layerWidget)
        # Add the button box
        self.layout.addWidget(self.buttonBox)
        debugger.print("Finished:: initialiser")

    def getLayer(self):
        """Return the edited layer.

        Parameters
        ----------
        None

        Returns
        -------
        self.layer : An instance of a single crystal layer

        """
        return self.layer

    def drawLayerWidget(self):
        """Create a layer widget showing all the information about the layer.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        debugger.print("drawLayerWidget")
        widget = QWidget()
        form = QFormLayout()
        label = QLabel("Layer type:")
        form.addRow(label,QLabel(self.message))
        label = QLabel("Material:")
        material = self.layer.getMaterial()
        materialName = material.getName()
        form.addRow(label,QLabel(materialName))
        label,layout = self.drawLayerWidgetLine1()
        form.addRow(label,layout)
        if self.layer.getMaterial().isTensor():
            label,layout = self.drawLayerWidgetLine2()
            form.addRow(label,layout)
            label,layout = self.drawLayerWidgetLine3()
            form.addRow(label,layout)
        widget.setLayout(form)
        return widget

    def drawLayerWidgetLine1(self):
        """Add the first line of the layer description.

        Shows the material name and the film thickness and units. 

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        hbox = QHBoxLayout()
        debugger.print("drawLayerWidgetLine1")
        # Define material thickness
        materialThickness = self.layer.getThickness()
        # Handle thickness
        film_thickness_sb = QDoubleSpinBox(self)
        film_thickness_sb.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Fixed)
        film_thickness_sb.setToolTip("Define the thin film thickness in the defined thickness units")
        film_thickness_sb.setRange(0,100000)
        film_thickness_sb.setSingleStep(0.01)
        film_thickness_sb.setValue(materialThickness)
        film_thickness_sb.valueChanged.connect(self.on_film_thickness_sb_changed)
        # Handle thickness units
        thicknessUnit = self.layer.getThicknessUnit()
        thickness_units_cb = QComboBox(self)
        thickness_units_cb.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Fixed)
        thickness_units_cb.setToolTip("Set the units to be used for thickness; either nm, um, mm or cm")
        thickness_units_cb.addItems( ["nm","um","mm","cm"] )
        index = thickness_units_cb.findText(thicknessUnit, Qt.MatchFixedString)
        thickness_units_cb.setCurrentIndex(index)
        thickness_units_cb.activated.connect(self.on_thickness_units_cb_activated)
        thicknessLabel = QLabel("Thickness:")
        thicknessLabel.setToolTip("Define the depth of the thin crystal in the defined thickness units.")
        # Create the line of widgets
        hbox.addWidget(film_thickness_sb)
        hbox.addWidget(thickness_units_cb)
        return thicknessLabel,hbox

    def drawLayerWidgetLine2(self):
        """Add the second line of the layer description.

        Shows the hkl specification and the azimuthal angle

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        debugger.print("drawLayerWidgetLine2")
        hbox = QHBoxLayout()
        # define hkl
        h_sb = QSpinBox(self)
        h_sb.setToolTip("Define the h dimension of the unique direction")
        h_sb.setRange(-20,20)
        h_sb.setSingleStep(1)
        h_sb.setValue(self.layer.getHKL()[0])
        h_sb.valueChanged.connect(lambda x: self.on_hkl_sb_changed(x,0))
        k_sb = QSpinBox(self)
        k_sb.setToolTip("Define the k dimension of the unique direction")
        k_sb.setRange(-20,20)
        k_sb.setSingleStep(1)
        k_sb.setValue(self.layer.getHKL()[1])
        k_sb.valueChanged.connect(lambda x: self.on_hkl_sb_changed(x,1))
        l_sb = QSpinBox(self)
        l_sb.setToolTip("Define the l dimension of the unique direction")
        l_sb.setRange(-20,20)
        l_sb.setSingleStep(1)
        l_sb.setValue(self.layer.getHKL()[2])
        l_sb.valueChanged.connect(lambda x: self.on_hkl_sb_changed(x,2))
        hklLabel = QLabel("hkl:")
        hklLabel.setToolTip("Define the crystal surface (hkl). Defines the unique direction in crystallographic units.")
        # define azimuthal angle
        azimuthal = self.layer.getAzimuthal()
        azimuthal_angle_sb = QDoubleSpinBox(self)
        azimuthal_angle_sb.setToolTip("Define the slab azimuthal angle (rotation of the crystal about the lab Z-axis).\nThe orientation of the crystal in the laboratory frame can be seen in the laboratory frame information below")
        azimuthal_angle_sb.setRange(-180,360)
        azimuthal_angle_sb.setSingleStep(10)
        azimuthal_angle_sb.setValue(azimuthal)
        azimuthal_angle_sb.valueChanged.connect(self.on_azimuthal_angle_sb_changed)
        azimuthalLabel = QLabel("Azimuthal:")
        # Create the line of widgets
        hbox.addWidget(h_sb)
        hbox.addWidget(k_sb)
        hbox.addWidget(l_sb)
        hbox.addWidget(azimuthalLabel)
        hbox.addWidget(azimuthal_angle_sb)
        return hklLabel,hbox

    def drawLayerWidgetLine3(self):
        """Add the third line of the layer description.

        Shows the relationship between crystal and labframe coordinate systems

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        debugger.print("drawLayerWidgetLine3")
        hbox = QHBoxLayout()
        label = QLabel("Lab frame\ninformation", self)
        label.setToolTip("The normal to the surface defines the Z-axis in the  lab frame\nThe incident and reflected light lie in the XZ plane\nThe p-polarization is direction lies in the XZ plane, s-polarisation is parallel to Y")
        self.labframe_w = QListWidget(self)
        fm = self.labframe_w.fontMetrics()
        h = fm.ascent() + fm.descent()
        self.labframe_w.setMaximumHeight(6*h)
        self.labframe_w.setToolTip("The normal to the surface defines the Z-axis in the  lab frame\nThe incident and reflected light lie in the XZ plane\nThe p-polarization is direction lies in the XZ plane, s-polarisation is parallel to Y")
        self.changeLabFrameInfo()
        hbox.addWidget(self.labframe_w)
        return label,hbox

    def changeLabFrameInfo(self):
        """Update the lab frame information in the labframe widget.

        Shows a, b and c crystal axis in the labframe.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        a,b,c = self.layer.getLabFrame()
        self.labframe_w.clear()
        self.labframe_w.addItem(f"crystal a-axis in lab frame: {a[0]: 3.5f}, {a[1]: 3.5f}, {a[2]: 3.5f}" )
        self.labframe_w.addItem(f"crystal b-axis in lab frame: {b[0]: 3.5f}, {b[1]: 3.5f}, {b[2]: 3.5f}" )
        self.labframe_w.addItem(f"crystal c-axis in lab frame: {c[0]: 3.5f}, {c[1]: 3.5f}, {c[2]: 3.5f}" )
        return

    def on_film_thickness_sb_changed(self,value):
        """Handle film thickness spin box change.

        Parameters
        ----------
        value : float
            The new thickness

        Returns
        -------
        None

        """
        debugger.print("on_film_thickness_sb_changed", value)
        self.layer.setThickness(value)
        return

    def on_azimuthal_angle_sb_changed(self,value):
        """Handle azimuthal spin box change.

        Parameters
        ----------
        value : float
            The new azimuthal angle in degrees

        Returns
        -------
        None

        """
        debugger.print("on_azimuthal_angl_sb_changed", value)
        self.layer.setAzimuthal(value)
        self.layer.changeLabFrameInfo()
        self.changeLabFrameInfo()
        return

    def on_hkl_sb_changed(self,value,hkorl):
        """Handle h, k, or l change.

        Parameters
        ----------
        value : int
           The value of h, k or l
        hkorl : int
            0, 1, or 2 for h, k or l respectively.

        Returns
        -------
        None

        """
        debugger.print("on_h_sb_changed", value)
        hkl = self.layer.getHKL()
        hkl[hkorl] = value
        self.layer.setHKL(hkl)
        self.layer.changeLabFrameInfo()
        self.changeLabFrameInfo()
        return

    def on_thickness_units_cb_activated(self, index):
        """Handle activation of thickness units combobox.

        Parameters
        ----------
        index : int
            The index of the selected item in the combobox.
            One of 'nm', 'um', 'mm', 'cm'

        Returns
        -------
        None

        """        
        debugger.print("Start:: on_thickness_units_cb_activated",index)
        units = ["nm","um","mm","cm"]
        unit = units[index]
        self.layer.setThicknessUnit(unit)
        return
